//////////////////////////////////////////////////////////////////////////
// BonSDK .NET binding interop library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.


#include "stdafx.h"

#include "ManagedLock.h"
#include "CreateBonInstance.h"
#include "DataQueue.h"
#include "StreamSink.h"
#include "MediaDecoderHandlerTrampoline.h"
#include "TuningSpaceInformation.h"
#include "EpgData.h"
#include "ProgramInformation.h"
#include "SelectedChannel.h"
#include "Tuner.h"

namespace Bon
{
	Tuner::Tuner(ITunerDefinition^ tunerDefinition, System::String^ casCardReader)
		: tunerDefinition_(tunerDefinition), lockCookie_(System::Guid::Empty), running_(false),
		enableEmmProcess_(false), enableLnbPower_(true), tuningSpaces_(nullptr),
		pTsSourceTuner_(0), pTsPacketSync_(0), pProgAnalyzer_(0), pTsDescrambler_(0), pStreamSink_(0),
		pTsSourceTunerTrampoline_(0), pTsPacketSyncTrampoline_(0), pProgAnalyzerTrampoline_(0), pTsDescramblerTrampoline_(0),
		packetSyncEvents_(gcnew System::Collections::Generic::List<IPacketSyncEvent^>()),
		descrambleEvents_(gcnew System::Collections::Generic::List<IDescrambleEvent^>())
	{
		try
		{
			pTsSourceTunerTrampoline_ = new Internal::MediaDecoderHandlerTrampoline();
			pTsSourceTunerTrampoline_->RegisterMediaDecoderEvent(
				gcnew Internal::MediaDecoderEventHandler(this, &Tuner::OnTsSourceTunerEvent));
			pTsPacketSyncTrampoline_ = new Internal::MediaDecoderHandlerTrampoline();
			pTsPacketSyncTrampoline_->RegisterMediaDecoderEvent(
				gcnew Internal::MediaDecoderEventHandler(this, &Tuner::OnTsPacketSyncEvent));
			pProgAnalyzerTrampoline_ = new Internal::MediaDecoderHandlerTrampoline();
			pProgAnalyzerTrampoline_->RegisterMediaDecoderEvent(
				gcnew Internal::MediaDecoderEventHandler(this, &Tuner::OnProgAnalyzerEvent));
			pTsDescramblerTrampoline_ = new Internal::MediaDecoderHandlerTrampoline();
			pTsDescramblerTrampoline_->RegisterMediaDecoderEvent(
				gcnew Internal::MediaDecoderEventHandler(this, &Tuner::OnTsDescramblerEvent));

			pTsSourceTuner_ = Internal::CreateBonInstance<ITsSourceTuner>(TEXT("CTsSourceTuner"),
				static_cast<CBonObject *>(pTsSourceTunerTrampoline_));
			pTsPacketSync_ = Internal::CreateBonInstance<ITsPacketSync>(TEXT("CTsPacketSync"),
				static_cast<CBonObject *>(pTsPacketSyncTrampoline_));
			pProgAnalyzer_ = Internal::CreateBonInstance<IProgAnalyzer>(TEXT("CProgAnalyzer"),
				static_cast<CBonObject *>(pProgAnalyzerTrampoline_));
			pTsDescrambler_ = Internal::CreateBonInstance<ITsDescrambler>(TEXT("CTsDescrambler"),
				static_cast<CBonObject *>(pTsDescramblerTrampoline_));
			pStreamSink_ = new Internal::StreamSink(pProgAnalyzer_, 512 * 1024);

			dynamic_cast<IMediaDecoder *>(pTsSourceTuner_)->SetOutputDecoder(dynamic_cast<IMediaDecoder *>(pTsPacketSync_));
			dynamic_cast<IMediaDecoder *>(pTsPacketSync_)->SetOutputDecoder(dynamic_cast<IMediaDecoder *>(pTsDescrambler_));
			dynamic_cast<IMediaDecoder *>(pTsDescrambler_)->SetOutputDecoder(dynamic_cast<IMediaDecoder *>(pProgAnalyzer_));
			dynamic_cast<IMediaDecoder *>(pProgAnalyzer_)->SetOutputDecoder(pStreamSink_);

			pin_ptr<const wchar_t> pTunerName = PtrToStringChars(tunerDefinition_->TunerName);

			int tunerResult = pTsSourceTuner_->OpenTuner(pTunerName);
			if (tunerResult > 0)
			{
				throw gcnew BonException("Cannot open tuner: Name={0}, Result={1}",
					tunerDefinition_->TunerName, tunerResult);
			}

			pin_ptr<const wchar_t> pCasCardReader = PtrToStringChars(casCardReader);

			if (pTsDescrambler_->OpenDescrambler(pCasCardReader) == false)
			{
				pTsSourceTuner_->CloseTuner();
				throw gcnew BonException("Cannot open card reader");
			}

			pTsPacketSync_->DiscardNullPacket(true);
			pTsDescrambler_->DiscardScramblePacket(false);
			pTsDescrambler_->EnableBuffering(true);
		}
		catch (...)
		{
			this->~Tuner();
			throw;
		}
	}

	Tuner::~Tuner()
	{
		Internal::ManagedLock lock(this);

		BON_SAFE_RELEASE(pTsSourceTuner_);
		BON_SAFE_RELEASE(pTsPacketSync_);
		BON_SAFE_RELEASE(pTsDescrambler_);
		BON_SAFE_RELEASE(pProgAnalyzer_);
		BON_SAFE_DELETE(pStreamSink_);
		BON_SAFE_DELETE(pTsDescramblerTrampoline_);
		BON_SAFE_DELETE(pProgAnalyzerTrampoline_);
		BON_SAFE_DELETE(pTsPacketSyncTrampoline_);
		BON_SAFE_DELETE(pTsSourceTunerTrampoline_);

		packetSyncEvents_->Clear();
		descrambleEvents_->Clear();
	}

	void Tuner::OnTsSourceTunerEvent(Internal::MediaDecoderEventArgs^ e)
	{
		BON_TRACE(TraceType_Debug, TEXT("Unknown OnTsSourceTunerEvent(%1!p!, %2!d!, %3!p!)"),
			e->MediaDecoder, e->EventID, e->Parameter);
	}

	void Tuner::OnTsPacketSyncEvent(Internal::MediaDecoderEventArgs^ e)
	{
		Internal::ManagedLock lock(packetSyncEvents_);

		switch (e->EventID)
		{
			case ITsPacketSync::EID_SYNC_ERR :
				for (int i = 0; i < packetSyncEvents_->Count; i++)
				{
					try
					{
						packetSyncEvents_[i]->OnSynchronizedError(static_cast<int>(e->Parameter));
					}
					catch (...)
					{
					}
				}
				break;
			
			case ITsPacketSync::EID_FORMAT_ERR :
				for (int i = 0; i < packetSyncEvents_->Count; i++)
				{
					try
					{
						packetSyncEvents_[i]->OnFormatError(static_cast<int>(e->Parameter));
					}
					catch (...)
					{
					}
				}
				break;

			case ITsPacketSync::EID_TRANSPORT_ERR :
				for (int i = 0; i < packetSyncEvents_->Count; i++)
				{
					try
					{
						packetSyncEvents_[i]->OnTransportError(static_cast<int>(e->Parameter));
					}
					catch (...)
					{
					}
				}
				break;

			case ITsPacketSync::EID_CONTINUITY_ERR :
				for (int i = 0; i < packetSyncEvents_->Count; i++)
				{
					try
					{
						packetSyncEvents_[i]->OnContinuityError(static_cast<int>(e->Parameter));
					}
					catch (...)
					{
					}
				}
				break;

			default:
				BON_TRACE(TraceType_Debug, TEXT("Unknown OnTsPacketSyncEvent(%1!p!, %2!d!, %3!p!)"),
					e->MediaDecoder, e->EventID, e->Parameter);
				break;
		}
	}

	void Tuner::OnTsDescramblerEvent(Internal::MediaDecoderEventArgs^ e)
	{
		Internal::ManagedLock lock(descrambleEvents_);

		switch (e->EventID)
		{
			case ITsDescrambler::EID_CANT_DESCRAMBLE :
				for (int i = 0; i < descrambleEvents_->Count; i++)
				{
					try
					{
						descrambleEvents_[i]->OnCannotDescramble(static_cast<int>(e->Parameter));
					}
					catch (...)
					{
					}
				}
				break;
			
			case ITsDescrambler::EID_ECM_PROCESSED :
				for (int i = 0; i < descrambleEvents_->Count; i++)
				{
					try
					{
						descrambleEvents_[i]->OnEcmProcessed(static_cast<DescramblingStates>(e->Parameter));
					}
					catch (...)
					{
					}
				}
				break;

			case ITsDescrambler::EID_EMM_PROCESSED :
				for (int i = 0; i < descrambleEvents_->Count; i++)
				{
					try
					{
						descrambleEvents_[i]->OnEmmProcessed(e->Parameter != 0);
					}
					catch (...)
					{
					}
				}
				break;

			default:
				BON_TRACE(TraceType_Debug, TEXT("Unknown OnTsDescramblerEvent(%1!p!, %2!d!, %3!p!)"),
					e->MediaDecoder, e->EventID, e->Parameter);
				break;
		}
	}

	void Tuner::OnProgAnalyzerEvent(Internal::MediaDecoderEventArgs^ e)
	{
		BON_TRACE(TraceType_Debug, TEXT("Unknown OnProgAnalyzerEvent(%1!p!, %2!d!, %3!p!)"),
			e->MediaDecoder, e->EventID, e->Parameter);
	}

	void Tuner::RegisterEvent(IPacketSyncEvent^ packetSync)
	{
		Internal::ManagedLock lock(packetSyncEvents_);
		packetSyncEvents_->Add(packetSync);
	}

	void Tuner::RegisterEvent(IDescrambleEvent^ descramble)
	{
		Internal::ManagedLock lock(descrambleEvents_);
		descrambleEvents_->Add(descramble);
	}

	void Tuner::UnregisterEvent(IPacketSyncEvent^ packetSync)
	{
		Internal::ManagedLock lock(packetSyncEvents_);
		packetSyncEvents_->Remove(packetSync);
	}

	void Tuner::UnregisterEvent(IDescrambleEvent^ descramble)
	{
		Internal::ManagedLock lock(descrambleEvents_);
		descrambleEvents_->Remove(descramble);
	}

	System::String^ Tuner::TunerName::get()
	{
		return tunerDefinition_->TunerName;
	}

	System::String^ Tuner::DeviceName::get()
	{
		return tunerDefinition_->DeviceName;
	}

	array<ITuningSpaceInformation^>^ Tuner::TuningSpaces::get()
	{
		return tunerDefinition_->TuningSpaces;
	}

	int Tuner::DeviceID::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetDeviceID();
	}

	bool Tuner::IsLocked::get()
	{
		return lockCookie_ != System::Guid::Empty;
	}

	void Tuner::Unlock(System::Guid cookie)
	{
		Internal::ManagedLock lock(this);

		if (cookie != lockCookie_)
		{
			throw gcnew System::InvalidOperationException("Invalid lock cookie.");
		}

		lockCookie_ = System::Guid::Empty;
	}

	bool Tuner::IsRunning::get()
	{
		return running_;
	}

	array<System::Byte>^ Tuner::CardID::get()
	{
		Internal::ManagedLock lock(this);

		CMediaData mediaData(0);
		pTsDescrambler_->GetCardID(&mediaData);

		array<System::Byte>^ buffer = gcnew array<System::Byte>(mediaData.GetSize());
		pin_ptr<BYTE> pBuffer = &buffer[0];
		memcpy(pBuffer, mediaData.GetData(), mediaData.GetSize());

		return buffer;
	}

	bool Tuner::EnableEmmProcess::get()
	{
		return enableEmmProcess_;
	}

	void Tuner::EnableEmmProcess::set(bool value)
	{
		Internal::ManagedLock lock(this);

		if (lockCookie_ != System::Guid::Empty)
		{
			throw gcnew System::InvalidOperationException("Tuner locked.");
		}

		if (enableEmmProcess_ != value)
		{
			enableEmmProcess_ = value;
			if (running_ == true)
			{
				pTsDescrambler_->EnableEmmProcess(enableEmmProcess_);
			}
		}
	}

	bool Tuner::EnableLnbPower::get()
	{
		return enableLnbPower_;
	}

	void Tuner::EnableLnbPower::set(bool value)
	{
		Internal::ManagedLock lock(this);

		if (lockCookie_ != System::Guid::Empty)
		{
			throw gcnew System::InvalidOperationException("Tuner locked.");
		}

		if (enableLnbPower_ != value)
		{
			enableLnbPower_ = value;
			if (running_ == true)
			{
				pTsSourceTuner_->SetLnbPower(enableLnbPower_);
			}
		}
	}

	float Tuner::SignalLevel::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetSignalLevel();
	}

	int Tuner::TuningSpace::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetCurSpace();
	}

	int Tuner::Channel::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetCurChannel();
	}

	ISelectedChannel^ Tuner::GetChannel()
	{
		Internal::ManagedLock lock(this);

		return gcnew SelectedChannel(pTsSourceTuner_->GetCurSpace(), pTsSourceTuner_->GetCurChannel());
	}

	void Tuner::SetChannel(int tuningSpace, int channel, int programID)
	{
		Internal::ManagedLock lock(this);

		if (lockCookie_ != System::Guid::Empty)
		{
			throw gcnew System::InvalidOperationException("Tuner locked.");
		}

		// ύXȂ̏ꍇ͖
		if ((pTsSourceTuner_->GetCurSpace() == tuningSpace) && (pTsSourceTuner_->GetCurChannel() == channel))
		{
			return;
		}

		pTsSourceTuner_->SetChannel(tuningSpace, channel);
		pStreamSink_->ChangeProgramIDForAllQueue(programID);
	}

	System::Guid Tuner::SetChannelWithLock(int tuningSpace, int channel, int programID)
	{
		Internal::ManagedLock lock(this);

		if (lockCookie_ != System::Guid::Empty)
		{
			throw gcnew System::InvalidOperationException("Tuner locked.");
		}

		// ύXꍇ
		if ((pTsSourceTuner_->GetCurSpace() != tuningSpace) || (pTsSourceTuner_->GetCurChannel() != channel))
		{
			pTsSourceTuner_->SetChannel(tuningSpace, channel);
			pStreamSink_->ChangeProgramIDForAllQueue(programID);
		}

		// bNNbL[蓖Ă
		lockCookie_ = System::Guid::NewGuid();
		return lockCookie_;
	}

	int Tuner::NetworkID::get()
	{
		Internal::ManagedLock lock(this);

		return pProgAnalyzer_->GetNetworkID();
	}

	System::String^ Tuner::NetworkName::get()
	{
		Internal::ManagedLock lock(this);

		const CBonString networkName = pProgAnalyzer_->GetNetworkName();
		return gcnew System::String(networkName);
	}

	int Tuner::TransportStreamID::get()
	{
		Internal::ManagedLock lock(this);

		return static_cast<int>(pProgAnalyzer_->GetTsID());
	}

	System::DateTime Tuner::StreamTime::get()
	{
		Internal::ManagedLock lock(this);

		SYSTEMTIME streamTime;
		memset(&streamTime, 0, sizeof streamTime);
		bool result = pProgAnalyzer_->GetStreamTime(&streamTime);
		if (result == false)
		{
			return System::DateTime::MinValue;
		}

		// GetStreamTime() is local time
		return System::DateTime(streamTime.wYear, streamTime.wMonth, streamTime.wDay,
			streamTime.wHour, streamTime.wMinute, streamTime.wSecond, streamTime.wMilliseconds,
			System::DateTimeKind::Local);
	}

	array<IProgramInformation^>^ Tuner::Programs::get()
	{
		System::Collections::Generic::List<IProgramInformation^>^ programs =
			gcnew System::Collections::Generic::List<IProgramInformation^>();

		Internal::ManagedLock lock(this);

		for (WORD i = 0; i < pProgAnalyzer_->GetProgramNum(); i++)
		{
			WORD programID = pProgAnalyzer_->GetProgramID(i);

			WORD programVideoID = pProgAnalyzer_->GetProgramVideoPID(programID);
			WORD programAudioID = pProgAnalyzer_->GetProgramAudioPID(programID);
			WORD programPcrID = pProgAnalyzer_->GetProgramPcrPID(programID);

			const CBonString serviceName = pProgAnalyzer_->GetServiceName(programID);

			const IEpgItem *pEpgItem = pProgAnalyzer_->GetCurrentEpgInfo(programID);

			programs->Add(gcnew ProgramInformation(
				programID, programVideoID, programAudioID, programPcrID,
				gcnew System::String(serviceName), gcnew EpgData(pEpgItem)));
		}

		return programs->ToArray();
	}

	DescramblingStates Tuner::DescramblingStatus::get()
	{
		Internal::ManagedLock lock(this);

		return static_cast<DescramblingStates>(
			pTsDescrambler_->GetDescramblingState(pProgAnalyzer_->GetCurrentProgramID()));
	}

	IDataQueue^ Tuner::CreateQueue(int programID, __int64 maxSize)
	{
		Internal::ManagedLock lock(this);

		System::Collections::Generic::KeyValuePair<IDataQueue^, int> queue =
			pStreamSink_->CreateQueue(this, programID, maxSize);
		if (queue.Value == 1)
		{
			pTsDescrambler_->EnableEmmProcess(enableEmmProcess_);
			pTsSourceTuner_->SetLnbPower(enableLnbPower_);
			dynamic_cast<IMediaDecoder *>(pTsSourceTuner_)->PlayDecoder();

			running_ = true;
		}
		return queue.Key;
	}

	void Tuner::ChangeProgramID(DataQueue^ queue, int newProgramID)
	{
		Internal::ManagedLock lock(this);

		pStreamSink_->ChangeProgramID(queue, newProgramID);
	}

	void Tuner::RemoveQueue(DataQueue^ queue)
	{
		Internal::ManagedLock lock(this);

		int count = pStreamSink_->RemoveQueue(queue);
		if (count == 0)
		{
			if (pTsSourceTuner_ != nullptr)
			{
				dynamic_cast<IMediaDecoder *>(pTsSourceTuner_)->StopDecoder();
				pTsDescrambler_->EnableEmmProcess(false);
				pTsSourceTuner_->SetLnbPower(false);

				running_ = false;
			}
		}
	}
}
