//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service remote controller
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Net;
using System.Net.NetworkInformation;
using System.Reflection;
using System.Runtime.InteropServices;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Bon.Remoting;
using System.IO;

namespace Bon
{
	/// <summary>
	/// [gRg[tH[łB
	/// </summary>
	public partial class MainForm : Form
	{
		#region Interop
		[Flags]
		private enum InsertMenuFlags : int
		{
			String = 0x000,
			ByPosition = 0x400,
			Separator = 0x800
		}

		private enum ProcessorArchitecture : ushort
		{
			INTEL = 0,
			MIPS = 1,
			ALPHA = 2,
			PPC = 3,
			SHX = 4,
			ARM = 5,
			IA64 = 6,
			ALPHA64 = 7,
			MSIL = 8,
			AMD64 = 9,
			IA32onWin64 = 10
		}
		
		[StructLayout(LayoutKind.Sequential)]
		private struct SYSTEM_INFO
		{
			public ProcessorArchitecture processorArchitecture;
			ushort reserved;
			public uint pageSize;
			public IntPtr minimumApplicationAddress;
			public IntPtr maximumApplicationAddress;
			public IntPtr activeProcessorMask;
			public uint numberOfProcessors;
			public uint processorType;
			public uint allocationGranularity;
			public ushort processorLevel;
			public ushort processorRevision;
		}

		[DllImport("user32.dll")]
		private static extern IntPtr GetSystemMenu(IntPtr hWnd, bool revert);
		[DllImport("user32.dll")]
		private static extern bool InsertMenu(IntPtr hMenu,
			int position, InsertMenuFlags flags, int idNewItem, string newItem);
		[DllImport("kernel32.dll")]
		private static extern void GetNativeSystemInfo(ref SYSTEM_INFO systemInfo);

		private const int WM_SYSCOMMAND = 0x0112;
		private const int IDM_ABOUT = 1000;
		#endregion

		private delegate void AddStreamServiceEntryDelegate(TreeNode streamServiceNode);
		private delegate void AddTunerEntryDelegate(
			TreeNode streamServiceNode, TreeNode tunerNode, IPEndPoint[] clientEndPoints);

		private static readonly IPEndPointComparer ipEndPointComparer_ = new IPEndPointComparer();

		private readonly Dictionary<IPAddress, IPAddress> localEndPointAddresses_ = new Dictionary<IPAddress, IPAddress>();
		private readonly IRemoteStreamService remoting_ = new RemoteStreamService();
		private bool baloonTipShown_ = false;
		private bool updating_ = false;
		private TreeNode currentTunerNode_;
		
		/// <summary>
		/// RXgN^łB
		/// </summary>
		/// <param name="minimized">ŏĕ\Ȃtrue</param>
		public MainForm(bool minimized)
		{
			InitializeComponent();

			if (minimized == true)
			{
				this.WindowState = FormWindowState.Minimized;
			}

			// VXej[ǉ
			IntPtr sysMenuHandle = GetSystemMenu(this.Handle, false);
			InsertMenu(sysMenuHandle, 5, InsertMenuFlags.ByPosition | InsertMenuFlags.Separator, 0, string.Empty);
			InsertMenu(sysMenuHandle, 6, InsertMenuFlags.ByPosition, IDM_ABOUT, "&About remote controller...");
			
			// tH[ĩTCY𒲐
			updating_ = true;
			this.Size = Properties.Settings.Default.MainFormSize;
			mainTreeView.Height = Properties.Settings.Default.MainTreeViewHeight;
			tabControl.Width = Properties.Settings.Default.TabControlWidth;
			updating_ = false;

			// SẴC^[tFCX񋓂
			foreach (NetworkInterface ni in NetworkInterface.GetAllNetworkInterfaces())
			{
				foreach (UnicastIPAddressInformation ip in ni.GetIPProperties().UnicastAddresses)
				{
					localEndPointAddresses_[ip.Address] = ip.Address;
				}
			}

			// L^ς݂̃Xg[T[rXURLΐ
			StringCollection addStreamServiceHistory = Properties.Settings.Default.AddStreamServiceHistory;
			if (addStreamServiceHistory == null)
			{
				addStreamServiceHistory = new StringCollection();
				Properties.Settings.Default.AddStreamServiceHistory = addStreamServiceHistory;
				Properties.Settings.Default.Save();
			}

			// L^ς݂̃Xg[T[rXURLQ擾
			StringCollection streamServiceBaseUrls = Properties.Settings.Default.StreamServiceBaseUrls;
			if (streamServiceBaseUrls == null)
			{
				streamServiceBaseUrls = new StringCollection();
				Properties.Settings.Default.StreamServiceBaseUrls = streamServiceBaseUrls;
				Properties.Settings.Default.Save();
			}

			// Xg[T[rXURLQAÑXg[T[rXڑ񋓂
			foreach (string streamServiceBaseUrl in streamServiceBaseUrls)
			{
				try
				{
					// ڑɎԂ邩Ȃ̂ŁAXbhv[ŏ
					AddRemoteTunerCollection(new Uri(streamServiceBaseUrl));
				}
				catch
				{
				}
			}
		}

		/// <summary>
		/// tH[B
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			remoting_.Dispose();
		}

		/// <summary>
		/// `[i[RNVǉ܂B
		/// </summary>
		/// <param name="streamServiceBaseUrl">Xg[T[rXx[XURL</param>
		/// <remarks>񓯊ŃXg[T[rXm[hǉ܂B</remarks>
		private void AddRemoteTunerCollection(Uri streamServiceBaseUrl)
		{
			TreeNode streamServiceNode = null;

			mainTreeView.BeginUpdate();
			try
			{
				// c[m[h𐶐
				streamServiceNode = new TreeNode(streamServiceBaseUrl.ToString());
				streamServiceNode.ImageIndex = 0;
				streamServiceNode.SelectedImageIndex = 0;
				streamServiceNode.Tag = new StreamServiceTag(streamServiceBaseUrl, remoting_);

				mainTreeView.Nodes.Add(streamServiceNode);
				mainTreeView.ExpandAll();
			}
			finally
			{
				mainTreeView.EndUpdate();
			}

			// ȂIԂɂ
			if (mainTreeView.Nodes.Count == 1)
			{
				mainTreeView.SelectedNode = streamServiceNode;
			}

			// ڑɎԂ邩Ȃ̂ŁAXbhv[ŏ
			ThreadPool.QueueUserWorkItem(this.GetRemoteTunerCollectionEntry, streamServiceNode);
		}

		/// <summary>
		/// `[i[RNVXV܂B
		/// </summary>
		/// <param name="streamServiceNode">XVXg[T[rXc[m[h</param>
		private void UpdateRemoteTunerCollection(TreeNode streamServiceNode)
		{
			StreamServiceTag tag = (StreamServiceTag)streamServiceNode.Tag;

			mainTreeView.BeginUpdate();
			try
			{
				streamServiceNode.ImageIndex = 0;
				streamServiceNode.SelectedImageIndex = 0;
				tag.Exception = null;
				streamServiceNode.Nodes.Clear();
			}
			finally
			{
				mainTreeView.EndUpdate();
			}

			// ڑɎԂ邩Ȃ̂ŁAXbhv[ŏ
			ThreadPool.QueueUserWorkItem(this.GetRemoteTunerCollectionEntry, streamServiceNode);
		}

		/// <summary>
		/// IRemoteTunerCollectionC^[tFCXɐڑXbhv[Gg|CgłB
		/// </summary>
		/// <param name="state">ڑTreeNodeNX̃CX^X</param>
		private void GetRemoteTunerCollectionEntry(object state)
		{
			TreeNode node = (TreeNode)state;
			StreamServiceTag tag = (StreamServiceTag)node.Tag;

			try
			{
				// RNV񋓂A`[i[Q
				foreach (IRemoteTuner tuner in tag.TunerCollection)
				{
					// `[i[݂̌̏Ԃ𓾂
					IRemoteTunerStatus status = tuner.GetTunerStatus();

					// `[i[ɑΉc[m[h𐶐
					TreeNode tunerNode = new TreeNode(string.Format("[{0}] {1}", status.DeviceID, status.DeviceName));
					tunerNode.ImageIndex = 2;
					tunerNode.SelectedImageIndex = 2;
					tunerNode.Tag = tuner;

					this.Invoke(new AddTunerEntryDelegate(this.AddTunerEntry), node, tunerNode, status.ClientEndPoints);
				}
			}
			catch (Exception ex)
			{
				tag.Exception = ex;
				this.Invoke(new AddStreamServiceEntryDelegate(this.SetStreamServiceInvalidEntry), node);
			}
		}

		/// <summary>
		/// c[r[Ƀ`[i[ǉ}[VOGg|CgłB
		/// </summary>
		/// <param name="streamServiceNode">Xg[T[rXc[m[h</param>
		/// <param name="tunerNode">ǉc[m[h</param>
		/// <param name="clientEndPoints">ǉNCAgGh|CgQ</param>
		private void AddTunerEntry(TreeNode streamServiceNode, TreeNode tunerNode, IPEndPoint[] clientEndPoints)
		{
			mainTreeView.BeginUpdate();
			try
			{
				streamServiceNode.Nodes.Add(tunerNode);

				SetStreamClients(tunerNode, clientEndPoints);

				streamServiceNode.Expand();
				tunerNode.Expand();

				UpdateWindow(mainTreeView.SelectedNode);
			}
			catch
			{
			}
			finally
			{
				mainTreeView.EndUpdate();
			}
		}

		/// <summary>
		/// Xg[T[rXɖACRݒ肷}[VOGg|CgłB
		/// </summary>
		/// <param name="streamServiceNode">ݒ肷c[m[h</param>
		private void SetStreamServiceInvalidEntry(TreeNode streamServiceNode)
		{
			try
			{
				streamServiceNode.ImageIndex = 1;
				streamServiceNode.SelectedImageIndex = 1;
			}
			catch
			{
			}
		}

		/// <summary>
		/// Gh|Cg擾܂B
		/// </summary>
		/// <param name="endPoint">Gh|Cg</param>
		/// <returns>Gh|Cg</returns>
		private static string GetEndPointString(IPEndPoint endPoint)
		{
			try
			{
				// DNSt݂
				IPHostEntry entry = Dns.GetHostEntry(endPoint.Address);
				return string.Format("{0}:{1}", entry.HostName, endPoint.Port);
			}
			catch
			{
				return endPoint.ToString();
			}
		}

		/// <summary>
		/// Xg[NCAgɑΉc[m[h𐶐܂B
		/// </summary>
		/// <param name="tunerNode">`[i[m[h</param>
		/// <param name="clientEndPoints">NCAgGh|Cg̔z</param>
		private void SetStreamClients(TreeNode tunerNode, IPEndPoint[] clientEndPoints)
		{
			TreeNode selectedNode = mainTreeView.SelectedNode;

			TreeNode otherClientNode = null;
			if (tunerNode.Nodes.Count == 0)
			{
				// ǈȊÕNCAgm[h͂̃m[hɊi[
				otherClientNode = new TreeNode("Other clients");
				otherClientNode.ImageIndex = 3;
				otherClientNode.SelectedImageIndex = 3;

				// I[ɎǈȊÕNCAgm[hǉ
				tunerNode.Nodes.Add(otherClientNode);
			}
			else
			{
				// I[̎ǈȊÕNCAgm[h擾
				otherClientNode = tunerNode.Nodes[tunerNode.Nodes.Count - 1];
			}

			// 
			SortedDictionary<IPEndPoint, TreeNode> nodeDictionary =
				new SortedDictionary<IPEndPoint, TreeNode>(ipEndPointComparer_);
			for (int i = 0; i < tunerNode.Nodes.Count; i++)
			{
				IPEndPoint endPoint = (IPEndPoint)tunerNode.Nodes[i].Tag;
				if (endPoint != null)
				{
					nodeDictionary.Add(endPoint, tunerNode.Nodes[i]);
				}
			}
			for (int i = 0; i < otherClientNode.Nodes.Count; i++)
			{
				IPEndPoint endPoint = (IPEndPoint)otherClientNode.Nodes[i].Tag;
				nodeDictionary.Add(endPoint, otherClientNode.Nodes[i]);
			}

			// Gh|Cg̏
			foreach (IPEndPoint clientEndPoint in clientEndPoints)
			{
				// ̃Gh|CgɊYm[hȂij
				if (nodeDictionary.ContainsKey(clientEndPoint) == false)
				{
					// Gh|CgɊYm[hǉ
					TreeNode node = new TreeNode(GetEndPointString(clientEndPoint));
					node.ImageIndex = 4;
					node.SelectedImageIndex = 4;
					node.Tag = clientEndPoint;
					if (localEndPointAddresses_.ContainsKey(clientEndPoint.Address) == true)
					{
						tunerNode.Nodes.Insert(tunerNode.Nodes.Count - 1, node);
					}
					else
					{
						otherClientNode.Nodes.Add(node);
					}
				}
				else
				{
					// ݂ꍇ͎폜
					nodeDictionary.Remove(clientEndPoint);
				}
			}

			// ɎcĂm[h͍폜Ώ
			foreach (TreeNode node in nodeDictionary.Values)
			{
				node.Remove();
			}

			// IȂ
			if (selectedNode != null)
			{
				if (selectedNode.TreeView != null)
				{
					mainTreeView.SelectedNode = selectedNode;
				}
			}
		}

		/// <summary>
		/// uǉv{^NbNB
		/// </summary>
		private void addStreamServiceButton_Click(object sender, EventArgs e)
		{
			// ߋɒǉUrl̗Xg擾
			StringCollection addStreamServiceHistory = Properties.Settings.Default.AddStreamServiceHistory;

			// Xg[T[rXǉtH[\
			using (AddStreamServiceForm form = new AddStreamServiceForm(addStreamServiceHistory))
			{
				if (form.ShowDialog(this) == DialogResult.OK)
				{
					// ݂̃Xg[T[rXURLQɖ
					Uri url = form.Url;
					if (Properties.Settings.Default.StreamServiceBaseUrls.IndexOf(url.ToString()) == -1)
					{
						// ǉ
						AddRemoteTunerCollection(url);

						// Xg[T[rXURLQɒǉ
						Properties.Settings.Default.StreamServiceBaseUrls.Add(url.ToString());
					}
					else
					{
						MessageBox.Show(this, "This URL is already registered.", this.Text,
							MessageBoxButtons.OK, MessageBoxIcon.Error);
					}

					// Xg̐擪ɑ}
					Properties.Settings.Default.AddStreamServiceHistory.Remove(url.ToString());
					Properties.Settings.Default.AddStreamServiceHistory.Insert(0, url.ToString());
					
					// ۑ
					Properties.Settings.Default.Save();
				}
			}
		}

		/// <summary>
		/// uR[_[\v{^NbNꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void viewRecorderToolStripButton_Click(object sender, EventArgs e)
		{
			StreamServiceTag tag = (StreamServiceTag)viewRecorderToolStripButton.Tag;

			List<StreamServiceTag> tags = new List<StreamServiceTag>();
			foreach (TreeNode node in mainTreeView.Nodes)
			{
				tags.Add((StreamServiceTag)node.Tag);
			}

			using (RecorderForm form = new RecorderForm(tag, tags))
			{
				form.ShowDialog(this);
			}
		}

		/// <summary>
		/// u^v{^NbNB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void recordNowToolStripButton_Click(object sender, EventArgs e)
		{
			KeyValuePair<IRemoteTuner, IPEndPoint> tunerInformation =
				(KeyValuePair<IRemoteTuner, IPEndPoint>)recordNowToolStripButton.Tag;
			IRemoteTuner tuner = tunerInformation.Key;
			IPEndPoint endPoint = tunerInformation.Value;
			IRemoteTunerStatus status = tunerInformation.Key.GetTunerStatus();

			// I\ȃR[_񋓂
			List<StreamServiceTag> tags = new List<StreamServiceTag>();
			foreach (TreeNode node in mainTreeView.Nodes)
			{
				StreamServiceTag tag = (StreamServiceTag)node.Tag;
				tags.Add(tag);
			}

			// `[i[NCAgIĂꍇ́ÃNCAgIǂĂvOIDgp
			int programID = (endPoint != null) ? tuner.GetProgramID(endPoint) : -1;

			// R[_[ItH[\
			using (SelectRecorderForm form = new SelectRecorderForm(
				status.TuningSpace, status.Channel, programID, tags, status.StreamUrls,
				Properties.Settings.Default.QuickRecordingFileName,
				Properties.Settings.Default.QuickRecordingStreamServiceBaseUrl,
				Properties.Settings.Default.QuickRecordingStorageID,
				Properties.Settings.Default.QuickRecordingStreamUrl))
			{
				if (form.ShowDialog(this) == DialogResult.OK)
				{
					form.SelectedRecorder.Recorder.StartRecord(
						form.StreamUrl, status.TuningSpace, status.Channel, programID,
						form.StorageID, form.FileName);

					Properties.Settings.Default.QuickRecordingFileName = form.FileName;
					Properties.Settings.Default.QuickRecordingStreamServiceBaseUrl = form.SelectedRecorder.StreamServiceBaseUrl;
					Properties.Settings.Default.QuickRecordingStorageID = form.StorageID;
					Properties.Settings.Default.QuickRecordingStreamUrl = form.StreamUrl;
					Properties.Settings.Default.Save();
				}
			}
		}

		/// <summary>
		/// utbVv{^ꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void refreshToolStripButton_Click(object sender, EventArgs e)
		{
			UpdateWindow(mainTreeView.SelectedNode);
		}

		/// <summary>
		/// uwvv{^ꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void helpToolStripButton_Click(object sender, EventArgs e)
		{
			Assembly assembly = Assembly.GetExecutingAssembly();
			string path = Path.Combine(Path.GetDirectoryName(assembly.Location), @"wwwroot\RemoteControllerHelp.html");
			Process.Start(path);
		}

		/// <summary>
		/// Cc[r[Ń}EX{^ꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e">}EX</param>
		private void mainTreeView_MouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button != MouseButtons.Right)
			{
				return;
			}

			TreeNode node = mainTreeView.GetNodeAt(e.Location);
			if (node == null)
			{
				return;
			}

			// ܂IԂƂ
			mainTreeView.SelectedNode = node;

			// ̃m[hXg[T[rXĂ
			if (node.Tag is StreamServiceTag)
			{
				// R[_[j[Lɂ
				StreamServiceTag tag = (StreamServiceTag)node.Tag;
				viewRecorderToolStripMenuItem.Enabled = tag.Exception == null;
				viewRecorderToolStripMenuItem.Tag = tag;

				// O΃j[Lɂ
				showExceptionToolStripMenuItem.Enabled = tag.Exception != null;
				
				// ReLXgj[\
				connectionContextMenuStrip.Show(mainTreeView, e.Location);
			}
			// ̃m[h`[i[Ă
			else if (node.Tag is IRemoteTuner)
			{
				// ReLXgj[\
				IRemoteTuner tuner = (IRemoteTuner)node.Tag;
				IRemoteTunerStatus status = tuner.GetTunerStatus();

				tunerContextMenuStrip.Items.Clear();
				foreach (Uri streamUrl in status.StreamUrls)
				{
					ToolStripMenuItem item = new ToolStripMenuItem(
						string.Format("Open player [{0}]", streamUrl.Host), treeNodeImageList.Images[14]);
					item.Click += new EventHandler(openPlayerToolStripMenuItem_Click);
					item.Tag = new KeyValuePair<IRemoteTuner, Uri>(tuner, streamUrl);
					tunerContextMenuStrip.Items.Add(item);
				}

				tunerContextMenuStrip.Show(mainTreeView, e.Location);
			}
			// ̃m[hNCAgĂ
			else if (node.Tag is IPEndPoint)
			{
				// ReLXgj[\
				clientContextMenuStrip.Show(mainTreeView, e.Location);
			}
		}

		/// <summary>
		/// R[_[\j[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void viewRecorderToolStripMenuItem_Click(object sender, EventArgs e)
		{
			StreamServiceTag tag = (StreamServiceTag)viewRecorderToolStripMenuItem.Tag;

			List<StreamServiceTag> tags = new List<StreamServiceTag>();
			foreach (TreeNode node in mainTreeView.Nodes)
			{
				tags.Add((StreamServiceTag)node.Tag);
			}

			using (RecorderForm form = new RecorderForm(tag, tags))
			{
				form.ShowDialog(this);
			}
		}

		/// <summary>
		/// Đڑj[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void reconnectToolStripMenuItem_Click(object sender, EventArgs e)
		{
			TreeNode node = mainTreeView.SelectedNode;
			if (node.Tag is StreamServiceTag)
			{
				UpdateRemoteTunerCollection(node);
			}
		}

		/// <summary>
		/// O\j[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void showExceptionToolStripMenuItem_Click(object sender, EventArgs e)
		{
			TreeNode node = mainTreeView.SelectedNode;
			if (node.Tag is StreamServiceTag)
			{
				StreamServiceTag tag = (StreamServiceTag)node.Tag;
				Exception ex = tag.Exception;
				if (ex != null)
				{
					using (ExceptionForm form = new ExceptionForm(ex))
					{
						form.ShowDialog(this);
					}
				}
			}
		}

		/// <summary>
		/// [uj[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void removeToolStripMenuItem_Click(object sender, EventArgs e)
		{
			TreeNode node = mainTreeView.SelectedNode;
			if (node.Tag is StreamServiceTag)
			{
				if (MessageBox.Show(this, "Remove this URL, OK?", this.Text,
					MessageBoxButtons.OKCancel, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2) == DialogResult.OK)
				{
					mainTreeView.Nodes.Remove(node);

					StreamServiceTag tag = (StreamServiceTag)node.Tag;
					Properties.Settings.Default.StreamServiceBaseUrls.Remove(tag.StreamServiceBaseUrl.ToString());
					Properties.Settings.Default.Save();
				}
				viewRecorderToolStripButton.Enabled = false;
			}
		}

		/// <summary>
		/// v[[JIꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void openPlayerToolStripMenuItem_Click(object sender, EventArgs e)
		{
			ToolStripMenuItem item = sender as ToolStripMenuItem;
			if (item == null)
			{
				return;
			}

			KeyValuePair<IRemoteTuner, Uri> entry = (KeyValuePair<IRemoteTuner, Uri>)item.Tag;

			// vC[pX`ĂȂ
			if (string.IsNullOrEmpty(Properties.Settings.Default.PlayerPath) == true)
			{
				// OS̃A[LeN`𓾂
				SYSTEM_INFO systemInfo = new SYSTEM_INFO();
				GetNativeSystemInfo(ref systemInfo);

				// A[LeN`pX𓾂
				string path = Environment.ExpandEnvironmentVariables(
					entry.Key.GetDefaultPlayerPath("Windows-" + systemInfo.processorArchitecture.ToString()));
				Process.Start(path, entry.Value.ToString());
			}
			// vC[pX`ĂȂ炻gp
			else
			{
				string path = Environment.ExpandEnvironmentVariables(Properties.Settings.Default.PlayerPath);
				Process.Start(path, entry.Value.ToString());
			}
		}

		/// <summary>
		/// NCAgؒfIꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void disconnectClientToolStripMenuItem_Click(object sender, EventArgs e)
		{
			TreeNode node = mainTreeView.SelectedNode;
			IPEndPoint endPoint = node.Tag as IPEndPoint;
			if (endPoint != null)
			{
				if (MessageBox.Show(this,
					string.Format("Force disconnect client [{0}], OK?", GetEndPointString(endPoint)), this.Text,
					MessageBoxButtons.OKCancel, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2) == DialogResult.OK)
				{
					TreeNode tunerNode = GetSelectedTunerNode(node);
					IRemoteTuner tuner = (IRemoteTuner)tunerNode.Tag;

					tuner.DisconnectStream(endPoint);
				}
			}
		}

		/// <summary>
		/// IĂ`[i[m[h擾܂B
		/// </summary>
		/// <param name="beginNode">TJnm[h</param>
		/// <returns>`[i[m[hIĂȂꍇnull</returns>
		private TreeNode GetSelectedTunerNode(TreeNode beginNode)
		{
			// Iꂽm[h΁AIRemoteTuner̃CX^X擾݂
			TreeNode selectedTunerNode = beginNode;
			while (selectedTunerNode != null)
			{
				if (selectedTunerNode.Tag is IRemoteTuner)
				{
					break;
				}
				selectedTunerNode = selectedTunerNode.Parent;
			}
			return selectedTunerNode;
		}

		/// <summary>
		/// EChEXV܂B
		/// </summary>
		/// <param name="selectedNode">Cc[r[̑IԂɂm[h</param>
		private void UpdateWindow(TreeNode selectedNode)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}

			updating_ = true;
			try
			{
				// `[i[m[hȉ̃m[hłȂ
				TreeNode selectedTunerNode = GetSelectedTunerNode(selectedNode);
				if (selectedTunerNode == null)
				{
					// ̃m[hXg[T[rXĂ
					if (selectedNode.Tag is StreamServiceTag)
					{
						// R[_[j[Lɂ
						StreamServiceTag tag = (StreamServiceTag)selectedNode.Tag;
						viewRecorderToolStripButton.Enabled = tag.Exception == null;
						viewRecorderToolStripButton.Tag = tag;
					}
					else
					{
						viewRecorderToolStripButton.Enabled = false;
					}

					// `ly[WύX
					SetChannelPage(null, null, null);

					// vpeBy[WύX
					SetTunerPropertyPage(null, null, null);

					// vOy[WύX
					SetProgramPage(null, null, null, null);

					// ^{^𖳌ɂ
					recordNowToolStripButton.Enabled = false;

					currentTunerNode_ = null;

					return;
				}
				else
				{
					// R[_[j[Lɂ
					StreamServiceTag tag = (StreamServiceTag)selectedTunerNode.Parent.Tag;
					viewRecorderToolStripButton.Enabled = tag.Exception == null;
					viewRecorderToolStripButton.Tag = tag;
				}

				// `[i[݂̌̏Ԃ𓾂
				IRemoteTuner tuner = (IRemoteTuner)selectedTunerNode.Tag;
				IRemoteTunerStatus status = tuner.GetTunerStatus();

				// Xg[NCAgQXV
				SetStreamClients(selectedTunerNode, status.ClientEndPoints);

				// `ly[WύX
				SetChannelPage(selectedTunerNode, tuner, status);

				// vpeBy[WύX
				SetTunerPropertyPage(selectedTunerNode, tuner, status);

				// Im[hXg[NCAgm[hȂ
				IPEndPoint endPoint = selectedNode.Tag as IPEndPoint;
				if (endPoint != null)
				{
					// vOy[WύX
					SetProgramPage(selectedTunerNode, selectedNode, tuner, endPoint);
				}
				else
				{
					// vOy[WύX
					SetProgramPage(selectedTunerNode, null, null, null);
				}

				// ^{^Lɂ
				recordNowToolStripButton.Tag = new KeyValuePair<IRemoteTuner, IPEndPoint>(tuner, endPoint);
				recordNowToolStripButton.Enabled = true;

				currentTunerNode_ = selectedTunerNode;
			}
			finally
			{
				updating_ = false;
			}
		}

		/// <summary>
		/// Cc[r[NbNꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void mainTreeView_MouseDown(object sender, MouseEventArgs e)
		{
			// m[hoȂΖ
			TreeNode selectedNode = mainTreeView.GetNodeAt(e.Location);
			if (selectedNode == null)
			{
				return;
			}
			UpdateWindow(selectedNode);
		}

		/// <summary>
		/// Cc[r[̃m[h삳悤ƂĂB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void mainTreeView_BeforeExpand(object sender, TreeViewCancelEventArgs e)
		{
			// WJłȂΖ
			if (e.Action != TreeViewAction.Expand)
			{
				return;
			}

			// ēh~
			if (updating_ == true)
			{
				return;
			}

			updating_ = true;
			try
			{
				// `[i[m[hȉ̃m[hłȂΖ
				TreeNode selectedNode = e.Node;
				TreeNode selectedTunerNode = GetSelectedTunerNode(selectedNode);
				if (selectedTunerNode == null)
				{
					return;
				}

				// `[i[݂̌̏Ԃ𓾂
				IRemoteTuner tuner = (IRemoteTuner)selectedTunerNode.Tag;
				IRemoteTunerStatus status = tuner.GetTunerStatus();

				// Xg[NCAgQXV
				SetStreamClients(selectedTunerNode, status.ClientEndPoints);
			}
			finally
			{
				updating_ = false;
			}
		}

		/// <summary>
		/// ^uy[WύXꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tabControl_SelectedIndexChanged(object sender, EventArgs e)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}

			updating_ = true;
			try
			{
				TreeNode selectedTunerNode = GetSelectedTunerNode(mainTreeView.SelectedNode);
				if (selectedTunerNode == null)
				{
					return;
				}
				IRemoteTuner tuner = (IRemoteTuner)selectedTunerNode.Tag;

				// `[i[݂̌̏Ԃ𓾂
				IRemoteTunerStatus status = (tuner != null) ? tuner.GetTunerStatus() : null;

				// vpeBy[WύX
				SetTunerPropertyPage(selectedTunerNode, tuner, status);
			}
			finally
			{
				updating_ = false;
			}
		}

		/// <summary>
		/// `lc[r[̃m[hIύXꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void channelTreeView_AfterSelect(object sender, TreeViewEventArgs e)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}

			TreeNode selectedTunerNode = GetSelectedTunerNode(mainTreeView.SelectedNode);
			if (selectedTunerNode == null)
			{
				return;
			}
			IRemoteTuner tuner = (IRemoteTuner)selectedTunerNode.Tag;

			updating_ = true;
			try
			{
				// `[i[̃`lύX
				TreeNode selectedChannelNode = channelTreeView.SelectedNode;
				if (selectedChannelNode.Tag is KeyValuePair<int, int>)
				{
					KeyValuePair<int, int> tag = (KeyValuePair<int, int>)selectedChannelNode.Tag;

					// `lύX̓bN̂߂Ɏs邩Ȃ
					IRemoteTunerStatus status = null;
					try
					{
						tuner.SetChannel(tag.Key, tag.Value, 0);
						status = tuner.GetTunerStatus();
					}
					catch (InvalidOperationException)
					{
						// `ly[W̍ĕ`
						status = tuner.GetTunerStatus();
						currentTunerNode_ = null;
						SetChannelPage(selectedTunerNode, tuner, status);
					}

					// Xg[NCAgQXV
					SetStreamClients(selectedTunerNode, status.ClientEndPoints);

					// vpeBy[WύX
					SetTunerPropertyPage(selectedTunerNode, tuner, status);

					// Cc[r[őIĂm[hXg[NCAgm[hȂ
					TreeNode selectedNode = mainTreeView.SelectedNode;
					if (selectedNode != null)
					{
						IPEndPoint endPoint = selectedNode.Tag as IPEndPoint;
						if (endPoint != null)
						{
							// vOy[WύX
							SetProgramPage(selectedTunerNode, selectedNode, tuner, endPoint);
						}
						else
						{
							// vOy[WύX
							SetProgramPage(selectedTunerNode, null, null, null);
						}
					}
				}
			}
			finally
			{
				updating_ = false;
			}
		}

		/// <summary>
		/// vOc[r[̃m[hIύXꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void programTreeView_AfterSelect(object sender, TreeViewEventArgs e)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}

			updating_ = true;
			try
			{
				TreeNode selectedNode = programTreeView.SelectedNode;
				while (selectedNode != null)
				{
					if (selectedNode.Tag is ProgramSetter)
					{
						break;
					}
					selectedNode = selectedNode.Parent;
				}
				if (selectedNode != null)
				{
					ProgramSetter programSetter = (ProgramSetter)selectedNode.Tag;
					programSetter.SetProgram();
				}
			}
			finally
			{
				updating_ = false;
			}
		}

		/// <summary>
		/// `ly[WύX܂B
		/// </summary>
		/// <param name="selectedTunerNode">IĂ`[i[m[h</param>
		/// <param name="tuner">`[i[̎Q</param>
		/// <param name="status">`[i[̏</param>
		private void SetChannelPage(TreeNode selectedTunerNode, IRemoteTuner tuner, IRemoteTunerStatus status)
		{
			// `[i[m[hɕύXȂΖ
			if (object.ReferenceEquals(selectedTunerNode, currentTunerNode_) == true)
			{
				return;
			}

			// XVJn
			channelTreeView.BeginUpdate();
			try
			{
				// ׂč폜čĎ擾
				channelTreeView.Nodes.Clear();

				// `[i[w肳Ă
				// i`[i[QƂ̎擾ɎsĂƁAnullƂȂj
				if ((tuner != null) && (status != null))
				{
					// `[jOXy[X擾
					TreeNode selectedChannelNode = null;
					IRemoteTuningSpaceInformation[] tuningSpaces = tuner.GetTuningSpaces();
					for (int i = 0; i < tuningSpaces.Length; i++)
					{
						// `[jOXy[XɃc[m[h𐶐
						TreeNode tuningSpaceNode = new TreeNode(tuningSpaces[i].Name);
						tuningSpaceNode.ImageIndex = 5;
						tuningSpaceNode.SelectedImageIndex = 5;
						tuningSpaceNode.Tag = new KeyValuePair<int, int>(i, 0);

						// ݂̑Ĩm[hƈv
						if (status.TuningSpace == i)
						{
							// ̃m[hIԂɂ
							selectedChannelNode = tuningSpaceNode;
						}

						// `l񋓂
						string[] channels = tuningSpaces[i].Channels;
						for (int j = 0; j < channels.Length; j++)
						{
							// `l̃c[m[h𐶐
							TreeNode channelNode = new TreeNode(channels[j]);
							channelNode.ImageIndex = 6;
							channelNode.SelectedImageIndex = 6;
							channelNode.Tag = new KeyValuePair<int, int>(i, j);

							// ݂̑Ĩm[hƈv
							if ((status.TuningSpace == i) && (status.Channel == j))
							{
								// ̃m[hIԂɂ
								selectedChannelNode = channelNode;
							}

							tuningSpaceNode.Nodes.Add(channelNode);
						}
						channelTreeView.Nodes.Add(tuningSpaceNode);
					}

					// ׂĕ\
					channelTreeView.ExpandAll();

					// `[i[ɐݒ肳Ăm[h
					if (selectedChannelNode != null)
					{
						// m[hIČ悤ɂ
						channelTreeView.SelectedNode = selectedChannelNode;
						selectedChannelNode.EnsureVisible();
					}
				}
			}
			finally
			{
				// XVI
				channelTreeView.EndUpdate();
			}
		}

		/// <summary>
		/// vpeBy[WύX܂B
		/// </summary>
		/// <param name="selectedTunerNode">IĂ`[i[m[h</param>
		/// <param name="tuner">`[i[̎Q</param>
		/// <param name="status">`[i[̏</param>
		private void SetTunerPropertyPage(TreeNode selectedTunerNode, IRemoteTuner tuner, IRemoteTunerStatus status)
		{
			// `[i[݂̌̏ԂXV
			propertyGrid.SelectedObject = status;
		}

		/// <summary>
		/// vOy[WύX܂B
		/// </summary>
		/// <param name="selectedTunerNode">IĂ`[i[m[h</param>
		/// <param name="selectedStreamClientNode">IĂXg[NCAgm[h</param>
		/// <param name="tuner">`[i[̎Q</param>
		/// <param name="clientEndPoint">Xg[NCAgGh|Cg</param>
		private void SetProgramPage(TreeNode selectedTunerNode, TreeNode selectedStreamClientNode, IRemoteTuner tuner, IPEndPoint clientEndPoint)
		{
			programTreeView.Nodes.Clear();

			// `[i[w肳Ă
			// i`[i[QƂ̎擾ɎsĂƁAnullƂȂj
			if (tuner != null)
			{
				int programID = tuner.GetProgramID(clientEndPoint);

				// vOc[𐶐
				TreeNode targetNode = null;
				foreach (IRemoteProgramInformation rpi in tuner.GetPrograms())
				{
					if ((rpi.EpgDataStatus & EpgDataStates.ShortInfoStored) == EpgDataStates.ShortInfoStored)
					{
						TreeNode node = new TreeNode(string.Format("[{0}] {1}", rpi.ProgramID, rpi.ServiceName));
						node.ImageIndex = 7;
						node.SelectedImageIndex = 7;
						node.Tag = new ProgramSetter(tuner, clientEndPoint, rpi.ProgramID);

						if ((rpi.EpgDataStatus & EpgDataStates.EventIDStored) == EpgDataStates.EventIDStored)
						{
							TreeNode eventTitleNode = new TreeNode(
								string.Format("Title: [{0}] {1}, {2}", rpi.EventID, rpi.EventTitle,
								rpi.IsScrambled ? "Scrambled" : "Not scrambled"));
							eventTitleNode.ImageIndex = 8;
							eventTitleNode.SelectedImageIndex = 8;
							node.Nodes.Add(eventTitleNode);

							TreeNode eventInfoNode = new TreeNode(
								string.Format("Information: {0}", rpi.EventInfo));
							eventInfoNode.ImageIndex = 9;
							eventInfoNode.SelectedImageIndex = 9;
							node.Nodes.Add(eventInfoNode);
						}

						if ((rpi.EpgDataStatus & EpgDataStates.ExtendInfoStored) == EpgDataStates.ExtendInfoStored)
						{
							TreeNode eventDetailNode = new TreeNode(
								string.Format("Detail: {0}", rpi.EventDetail));
							eventDetailNode.ImageIndex = 9;
							eventDetailNode.SelectedImageIndex = 9;
							node.Nodes.Add(eventDetailNode);
						}

						if ((rpi.EpgDataStatus & EpgDataStates.EventIDStored) == EpgDataStates.EventIDStored)
						{
							TreeNode timeNode = new TreeNode(
								string.Format("Start time: {0:yyyy/MM/dd HH:mm:ss}, Duration: {1}",
								rpi.StartTime, rpi.Duration));
							timeNode.ImageIndex = 10;
							timeNode.SelectedImageIndex = 10;
							node.Nodes.Add(timeNode);
						}

						if ((rpi.EpgDataStatus & EpgDataStates.VideoInfoStored) == EpgDataStates.VideoInfoStored)
						{
							TreeNode videoNode = new TreeNode(
								string.Format("Video: [{0}] {1}, {2}, {3}, {4}",
								rpi.ProgramVideoID, rpi.VideoResolution, rpi.IsVideoProgressive ? "Progressive" : "Interlace",
								rpi.IsVideoAspectWide ? "Wide aspect" : "Standard aspect", rpi.IsVideoPanAndScan ? "Pan and scan" : "Normal"));
							videoNode.ImageIndex = 11;
							videoNode.SelectedImageIndex = 11;
							node.Nodes.Add(videoNode);
						}

						if ((rpi.EpgDataStatus & EpgDataStates.AudioInfoStored) == EpgDataStates.AudioInfoStored)
						{
							TreeNode audioNode = new TreeNode(
								string.Format("Audio: [{0}] {1}",
								rpi.ProgramAudioID, rpi.AudioMode));
							audioNode.ImageIndex = 12;
							audioNode.SelectedImageIndex = 12;
							node.Nodes.Add(audioNode);
						}

						programTreeView.Nodes.Add(node);

						if (rpi.ProgramID == programID)
						{
							targetNode = node;
						}
					}
				}

				// 擾łĂȂꍇ
				if (programTreeView.Nodes.Count == 0)
				{
					TreeNode node = new TreeNode("(Capturing now...)");
					node.ImageIndex = 7;
					node.SelectedImageIndex = 7;
					programTreeView.Nodes.Add(node);
				}

				programTreeView.ExpandAll();

				// Iǒ̃vOm[hIԂɂiȂꍇ͎IǏԁj
				if (targetNode != null)
				{
					programTreeView.SelectedNode = targetNode;
				}
			}
		}

		/// <summary>
		/// Cc[r[̃TCYύXꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void mainTreeView_SizeChanged(object sender, EventArgs e)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}
			Properties.Settings.Default.MainTreeViewHeight = mainTreeView.Height;
			Properties.Settings.Default.Save();
		}

		/// <summary>
		/// ^uRg[̃TCYύXꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tabControl_SizeChanged(object sender, EventArgs e)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}
			Properties.Settings.Default.TabControlWidth = tabControl.Width;
			Properties.Settings.Default.Save();
		}

		/// <summary>
		/// CtH[̃TCYύXꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_SizeChanged(object sender, EventArgs e)
		{
			// ēh~
			if (updating_ == true)
			{
				return;
			}
			Properties.Settings.Default.MainFormSize = this.Size;
			Properties.Settings.Default.Save();
		}

		/// <summary>
		/// CtH[悤ƂĂB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// NotifyIcon\Ă΁AtH[N[Y𒆒fAPɌȂ
			if (notifyIcon.Visible == true)
			{
				e.Cancel = true;
				this.Visible = false;

				// 񂾂o[`bv\
				if (baloonTipShown_ == false)
				{
					notifyIcon.ShowBalloonTip(10000);
					baloonTipShown_ = true;
				}
			}
		}

		/// <summary>
		/// NotifyIcon_uNbNꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void notifyIcon_DoubleClick(object sender, EventArgs e)
		{
			this.Visible = true;
			if (this.WindowState == FormWindowState.Minimized)
			{
				this.WindowState = FormWindowState.Normal;
			}
			this.Activate();
		}

		/// <summary>
		/// I[vj[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void openRemoteControllerToolStripMenuItem_Click(object sender, EventArgs e)
		{
			this.Visible = true;
			if (this.WindowState == FormWindowState.Minimized)
			{
				this.WindowState = FormWindowState.Normal;
			}
			this.Activate();
		}

		/// <summary>
		/// Ij[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void exitRemoteControllerToolStripMenuItem_Click(object sender, EventArgs e)
		{
			// NotifyIconȂ邱ƂŁAClosingCxgŃCtH[鎖ł
			notifyIcon.Visible = false;
			this.Close();
		}

		/// <summary>
		/// Aboutj[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void aboutRemoteControllerToolStripMenuItem_Click(object sender, EventArgs e)
		{
			using (AboutForm form = new AboutForm())
			{
				form.ShowDialog(this);
			}
		}

		/// <summary>
		/// VXej[𐧌䂷邽߂̃EChEvV[W̃I[o[ChłB
		/// </summary>
		/// <param name="message">EChEbZ[W</param>
		protected override void WndProc(ref Message message)
		{
			if (message.Msg == WM_SYSCOMMAND)
			{
				switch (message.WParam.ToInt32())
				{
					case IDM_ABOUT:
						using (AboutForm form = new AboutForm())
						{
							form.ShowDialog(this);
						}
						return;
					default:
						break;
				}
			}
			base.WndProc(ref message);
		}

		/// <summary>
		/// IPEndPoint̔rZqłB
		/// </summary>
		private sealed class IPEndPointComparer : IComparer<IPEndPoint>
		{
			/// <summary>
			/// rs܂B
			/// </summary>
			/// <param name="x">rΏ1</param>
			/// <param name="y">rΏ2</param>
			/// <returns>ʂŎl</returns>
			public int Compare(IPEndPoint x, IPEndPoint y)
			{
				return x.ToString().CompareTo(y.ToString());
			}
		}
	}

	/// <summary>
	/// Xg[T[rXւ̎QƂƗOێNXłB
	/// </summary>
	internal sealed class StreamServiceTag
	{
		private readonly Uri streamServiceBaseUrl_;
		private readonly IRemoteTunerCollection tunerCollection_;
		private readonly IRemoteRecorder recorder_;
		private Exception exception_;

		/// <summary>
		/// RXgN^łB
		/// </summary>
		/// <param name="streamServiceBaseUrl">Xg[T[rXx[XURL</param>
		/// <param name="remoting">[gQƂ擾C^[tFCX</param>
		public StreamServiceTag(Uri streamServiceBaseUrl, IRemoteStreamService remoting)
		{
			streamServiceBaseUrl_ = streamServiceBaseUrl;
			tunerCollection_ = remoting.GetTunerCollection(streamServiceBaseUrl);
			recorder_ = remoting.GetRecorder(streamServiceBaseUrl);
		}

		/// <summary>
		/// Xg[T[rXx[XURL擾܂B
		/// </summary>
		public Uri StreamServiceBaseUrl
		{
			get
			{
				return streamServiceBaseUrl_;
			}
		}

		/// <summary>
		/// `[i[RNVւ̃[gQƂ擾܂B
		/// </summary>
		public IRemoteTunerCollection TunerCollection
		{
			get
			{
				return tunerCollection_;
			}
		}

		/// <summary>
		/// R[_[ւ̃[gQƂ擾܂B
		/// </summary>
		public IRemoteRecorder Recorder
		{
			get
			{
				return recorder_;
			}
		}

		/// <summary>
		/// O擾Eݒ肵܂B
		/// </summary>
		public Exception Exception
		{
			get
			{
				return exception_;
			}
			set
			{
				exception_ = value;
			}
		}
	}
}
