//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace Bon
{
	/// <summary>
	/// T[rX̐ݒȂǂstH[łB
	/// </summary>
	public partial class SettingForm : Form
	{
		private readonly StreamServiceRegistryAccessor accessor_ =
			new StreamServiceRegistryAccessor(true, CultureInfo.InvariantCulture);
		private readonly ServiceController controler_ =
			new ServiceController(StreamService.ServiceName);
		private bool baloonTipShown_ = false;
		private ServiceControllerStatus controllerStatus_ = ServiceControllerStatus.Stopped;

		public SettingForm()
		{
			controllerStatus_ = controler_.Status;

			InitializeComponent();

			tabControlImageList.Images.Add(SystemIcons.Information.ToBitmap());

			imageList.Images.Add(SystemIcons.Error.ToBitmap());
			imageList.Images.Add(SystemIcons.Warning.ToBitmap());
			imageList.Images.Add(SystemIcons.Information.ToBitmap());
			imageList.Images.Add(SystemIcons.Information.ToBitmap());
			imageList.Images.Add(SystemIcons.Error.ToBitmap());
			imageList.Images.Add(SystemIcons.Question.ToBitmap());

			UpdateButtons(controllerStatus_);

			propertyGrid.SelectedObject = accessor_;

			timer.Start();
		}

		private void SettingForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			timer.Stop();
			accessor_.Dispose();
			controler_.Dispose();
		}

		private void UpdateButtons(ServiceControllerStatus status)
		{
			switch (status)
			{
				case ServiceControllerStatus.Stopped:
					startToolStripButton.Enabled = true;
					stopToolStripButton.Enabled = false;
					restartToolStripButton.Enabled = false;
					statusToolStripButton.Enabled = false;
					break;
				case ServiceControllerStatus.Running:
					startToolStripButton.Enabled = false;
					stopToolStripButton.Enabled = true;
					restartToolStripButton.Enabled = true;
					statusToolStripButton.Enabled = true;
					break;
				default:
					startToolStripButton.Enabled = false;
					stopToolStripButton.Enabled = false;
					restartToolStripButton.Enabled = false;
					statusToolStripButton.Enabled = false;
					break;
			}
		}

		private void UpdateButtons()
		{
			controler_.Refresh();

			if (controler_.Status != controllerStatus_)
			{
				controllerStatus_ = controler_.Status;

				UpdateButtons(controllerStatus_);
				UpdateEventLog();
			}
		}

		private void startButton_Click(object sender, EventArgs e)
		{
			this.Cursor = Cursors.WaitCursor;
			try
			{
				controler_.Start();
			}
			finally
			{
				UpdateButtons();
				this.Cursor = Cursors.Default;
			}
		}

		private void stopButton_Click(object sender, EventArgs e)
		{
			this.Cursor = Cursors.WaitCursor;
			try
			{
				controler_.Stop();
			}
			finally
			{
				UpdateButtons();
				this.Cursor = Cursors.Default;
			}
		}

		private void restartButton_Click(object sender, EventArgs e)
		{
			this.Cursor = Cursors.WaitCursor;
			try
			{
				controler_.Stop();
				UpdateButtons();
				controler_.WaitForStatus(ServiceControllerStatus.Stopped, new TimeSpan(0, 0, 30));

				controler_.Start();
				controler_.WaitForStatus(ServiceControllerStatus.Running, new TimeSpan(0, 0, 30));
			}
			finally
			{
				UpdateButtons();
				this.Cursor = Cursors.Default;
			}
		}

		private void statusToolStripButton_Click(object sender, EventArgs e)
		{
			string path = string.Format("http://localhost:{0}{1}/", accessor_.StreamPort, accessor_.StreamBasePath);
			Process.Start(path);
		}

		private void helpToolStripButton_Click(object sender, EventArgs e)
		{
			Assembly assembly = Assembly.GetExecutingAssembly();
			string path = Path.Combine(Path.GetDirectoryName(assembly.Location), @"wwwroot\StreamServiceHelp.html");
			Process.Start(path);
		}

		private void timer_Tick(object sender, EventArgs e)
		{
			try
			{
				this.Invoke(new ThreadStart(UpdateButtons));
			}
			catch
			{
			}
		}

		private void UpdateEventLog()
		{
			// EventLogȂ
			if (tabControl.SelectedIndex == 1)
			{
				eventLogListView.BeginUpdate();
				try
				{
					using (EventLog eventLog = new EventLog("Application"))
					{
						EventLogEntryCollection entries = eventLog.Entries;

						eventLogListView.Items.Clear();

						for (int i = entries.Count - 1; i >= 0; i--)
						{
							EventLogEntry entry = entries[i];
							if (entry.Source == StreamService.ServiceName)
							{
								ListViewItem item = new ListViewItem();

								switch (entry.EntryType)
								{
									case EventLogEntryType.Error:
										item.Text = "Error";
										item.ImageIndex = 0;
										break;
									case EventLogEntryType.Warning:
										item.Text = "Warning";
										item.ImageIndex = 1;
										break;
									case EventLogEntryType.Information:
										item.Text = "Info";
										item.ImageIndex = 2;
										break;
									case EventLogEntryType.SuccessAudit:
										item.Text = "Success";
										item.ImageIndex = 3;
										break;
									case EventLogEntryType.FailureAudit:
										item.Text = "Failure";
										item.ImageIndex = 4;
										break;
									default:
										item.Text = "Unknown";
										item.ImageIndex = 5;
										break;
								}
								item.SubItems.Add(entry.TimeGenerated.ToString("yyyy/MM/dd HH:mm:ss"));
								item.SubItems.Add(entry.Message);
								item.Tag = entry;
								eventLogListView.Items.Add(item);
							}
						}
					}
				}
				finally
				{
					eventLogListView.EndUpdate();
				}
			}
		}

		private void tabControl_TabIndexChanged(object sender, EventArgs e)
		{
			UpdateEventLog();
		}

		private void eventLogListView_DoubleClick(object sender, EventArgs e)
		{
			if (eventLogListView.SelectedItems.Count == 0)
			{
				return;
			}

			EventLogEntry entry = (EventLogEntry)eventLogListView.SelectedItems[0].Tag;
			using (EventLogDetailForm form = new EventLogDetailForm(entry))
			{
				form.ShowDialog(this);
			}
		}

		private void SettingForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// NotifyIcon\Ă΁AtH[N[Y𒆒fAPɌȂ
			if (notifyIcon.Visible == true)
			{
				e.Cancel = true;
				this.Visible = false;

				// 񂾂o[`bv\
				if (baloonTipShown_ == false)
				{
					notifyIcon.ShowBalloonTip(10000);
					baloonTipShown_ = true;
				}
			}
		}

		private void notifyIcon_DoubleClick(object sender, EventArgs e)
		{
			this.Visible = true;
			if (this.WindowState == FormWindowState.Minimized)
			{
				this.WindowState = FormWindowState.Normal;
			}
			this.Activate();
		}

		/// <summary>
		/// I[vj[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void openSettingsToolStripMenuItem_Click(object sender, EventArgs e)
		{
			this.Visible = true;
			if (this.WindowState == FormWindowState.Minimized)
			{
				this.WindowState = FormWindowState.Normal;
			}
			this.Activate();
		}

		/// <summary>
		/// Ij[IꂽB
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void exitSettingsToolStripMenuItem_Click(object sender, EventArgs e)
		{
			// NotifyIconȂ邱ƂŁAClosingCxgŃCtH[鎖ł
			notifyIcon.Visible = false;
			this.Close();
		}
	}
}