//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Collections;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.Net;
using System.Net.NetworkInformation;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Http;
using System.Text;
using System.Threading;
using Microsoft.Win32;
using Bon.Remoting;

namespace Bon
{
	/// <summary>
	/// Xg[~OsNXłB
	/// </summary>
	/// <remarks>[eBOZbgAbvAXg[~OXbh𐧌䂵܂B</remarks>
	public sealed class StreamManager : IDisposable
	{
		private static readonly char[] invalidPathChars_;

		private readonly string[] externalAddresses_;
		private readonly Uri remoteControlUrl_;
		private readonly string streamBasePath_ = "/stream";
		private readonly string streamContentType_ = "video/mpeg";
		private HttpListener listener_;
		private HttpChannel channel_;
		private readonly Dictionary<string, ITuner> tuners_ = new Dictionary<string, ITuner>();
		private readonly Dictionary<string, TunerFacade> tunerFacades_ = new Dictionary<string, TunerFacade>();
		private TunerCollectionFacade remoteTuners_;
		private RecorderManager recorderManager_;
		private readonly bool enableRecorder_ = true;
		private readonly bool enableLocalWebServer_ = true;
		private readonly string localWebServerRootPath_ =
			StreamServiceRegistryAccessor.GetDefaultLocalWebServerRootPath();

		/// <summary>
		/// ^CvCjVCUłB
		/// </summary>
		static StreamManager()
		{
			// pXQ𐶐
			List<char> invalidPathChars = new List<char>(new char[]
			{
				'\\', ':', ';', '&', '~', '#', '@', '<', '>', '"',
				'!', '?', '+', '*', '%', '$', '+', '^', '|', '`', '\''
			});
			for (int i = 0; i < 0x20; i++)
			{
				invalidPathChars.Add((char)i);
			}
			invalidPathChars_ = invalidPathChars.ToArray();
		}

		/// <summary>
		/// RXgN^łB
		/// </summary>
		public StreamManager()
		{
			int streamPort = 26110;
			string streamNameFormat = "{0}{1}.ts";
			int remotePort = 26111;
			string remoteBasePath = "/remote";
			long minimumSpaceRequirements = 100 * 1024 * 1024;
			int bufferSize = 4 * 1024 * 1024;
			string databasePath = StreamServiceRegistryAccessor.GetDefaultDatabasePath();
			TimeSpan extraTimeSpan = StreamServiceRegistryAccessor.GetDefaultExtraTimeSpan();
			Dictionary<string, string> storagePaths = new Dictionary<string, string>();
			Dictionary<string, string> defaultPlayers =
				new Dictionary<string, string>(StreamServiceRegistryAccessor.GetDefaultDefaultPlayers());

			// WXg擾
			try
			{
				using (StreamServiceRegistryAccessor accessor =
					new StreamServiceRegistryAccessor(false, CultureInfo.InvariantCulture))
				{
					streamPort = accessor.StreamPort;
					streamBasePath_ = accessor.StreamBasePath;
					streamNameFormat = accessor.StreamNameFormat;
					streamContentType_ = accessor.StreamContentType;
					remotePort = accessor.RemotePort;
					remoteBasePath = accessor.RemoteBasePath;
					enableLocalWebServer_ = accessor.EnableLocalWebServer;
					localWebServerRootPath_ = accessor.LocalWebServerRootPath;
					enableRecorder_ = accessor.EnableRecorder;
					minimumSpaceRequirements = accessor.MinimumSpaceRequirements;
					bufferSize = accessor.BufferSize;
					databasePath = accessor.DatabasePath;
					extraTimeSpan = accessor.ExtraTimeSpan;

					using (RegistryDictionary rawStoragePaths = accessor.StoragePaths)
					{
						foreach (KeyValuePair<string, string> entry in rawStoragePaths)
						{
							storagePaths[entry.Key] = entry.Value;
						}
					}

					using (RegistryDictionary rawDefaultPlayer = accessor.DefaultPlayers)
					{
						foreach (KeyValuePair<string, string> entry in rawDefaultPlayer)
						{
							defaultPlayers[entry.Key] = entry.Value;
						}
					}
				}
			}
			catch
			{
			}

			// URL̃`FbN
			if (streamBasePath_.StartsWith("/") == false)
			{
				throw new BonException(
					"StreamBasePath is not start with \"/\": StreamBasePath=\"{0}\"", streamBasePath_);
			}
			if (streamBasePath_.StartsWith("/") == false)
			{
				throw new BonException(
					"RemoteBasePath is not start with \"/\": RemoteBasePath=\"{0}\"", streamBasePath_);
			}

			// SẴlbg[NC^[tFCX񋓂
			List<string> temp = new List<string>();
			foreach (NetworkInterface ni in NetworkInterface.GetAllNetworkInterfaces())
			{
				foreach (UnicastIPAddressInformation ip in ni.GetIPProperties().UnicastAddresses)
				{
					if (IPAddress.IsLoopback(ip.Address) == false)
					{
						try
						{
							IPHostEntry entry = Dns.GetHostEntry(ip.Address);
							temp.Add(entry.HostName);
						}
						catch
						{
							temp.Add(ip.Address.ToString());
						}
					}
				}
			}
			if (temp.Count == 0)
			{
				temp.Add(Environment.MachineName.ToLower());
			}
			externalAddresses_ = temp.ToArray();

			// HTTPXi𐶐
			listener_ = new HttpListener();
			listener_.Prefixes.Add(string.Format("http://+:{0}{1}/", streamPort, streamBasePath_));
			
			// [eBO`l𐶐
			Dictionary<string, string> properties = new Dictionary<string, string>();
			properties.Add("port", remotePort.ToString());
			channel_ = new HttpChannel(properties, null, null);
			ChannelServices.RegisterChannel(channel_, false);
			remoteControlUrl_ = new Uri(
				string.Format("http://{0}:{1}{2}/", externalAddresses_[0], remotePort, remoteBasePath));

			// `[i[񋓂āAXg[~Oƃ[eBȌs
			StringBuilder sb = new StringBuilder();
			List<IRemoteTuner> remoteTuners = new List<IRemoteTuner>();
			TunerFactory tunerFactory = new TunerFactory();
			foreach (ITunerDefinition tunerDefinition in tunerFactory)
			{
				try
				{
					while (true)
					{
						// `[i[񋓂
						ITuner tuner = tunerDefinition.CreateTuner(tunerDefinition.DefaultCasCardReader);

						// "http://+:26110/stream/CBonBlackTuner0.ts"
						string path = string.Format("{0}/{1}", streamBasePath_,
							string.Format(streamNameFormat, tuner.TunerName, tuner.DeviceID));
						tuners_.Add(path, tuner);

						TunerFacade tunerFacade = new TunerFacade(
							tuner, externalAddresses_, streamPort, path, defaultPlayers);
						tunerFacades_.Add(path, tunerFacade);

						// "http://+:26111/remote/CBonBlackTuner0.rem"
						string uri = string.Format("{0}/{1}{2}.rem",
							remoteBasePath, tuner.TunerName, tuner.DeviceID);
						RemotingServices.Marshal(tunerFacade, uri.TrimStart('/'), typeof(IRemoteTuner));
						remoteTuners.Add(tunerFacade);

						sb.AppendFormat("\r\nDevice name: {0} [{1}]\r\nStream path: {2}\r\n",
							tuner.DeviceName, tuner.DeviceID, path);
					}
				}
				// G[`[i[͖
				catch (BonException ex)
				{
					Logger.Write(TraceType.Debug, ex);
				}
			}

			// `[i[ȏ㌩
			if (tuners_.Count >= 1)
			{
				// `[i[RNVJ
				remoteTuners_ = new TunerCollectionFacade(remoteTuners.ToArray());
				string tunerCollectionPath = string.Format("{0}/{1}", remoteBasePath, RemoteStreamService.TunerCollectionUri);
				RemotingServices.Marshal(remoteTuners_, tunerCollectionPath.TrimStart('/'), typeof(IRemoteTunerCollection));
				sb.AppendFormat("\r\nTuner collection path: {0}\r\n", tunerCollectionPath);
			}
			else
			{
				sb.Append("\r\nTuner not found.\r\n");
			}

			// R[_[J
			recorderManager_ = new RecorderManager(
				databasePath, storagePaths, minimumSpaceRequirements, bufferSize, extraTimeSpan);
			string recorderPath = string.Format("{0}/{1}", remoteBasePath, RemoteStreamService.RecorderUri);
			RemotingServices.Marshal(recorderManager_, recorderPath.TrimStart('/'), typeof(IRemoteRecorder));
			sb.AppendFormat("\r\nRecorder path: {0} ({1}, Count={2})\r\n", recorderPath,
				enableRecorder_ ? "Enable" : "Disable", recorderManager_.Count);

			// [JEFuT[oL
			if (enableLocalWebServer_ == true)
			{
				sb.AppendFormat("\r\nEnable local web server: {0}\r\n", localWebServerRootPath_);
			}

			Logger.Write("{0}\r\n{1}", this.GetType().Assembly.FullName, sb.ToString());
		}

		/// <summary>
		/// Dispose\bhłB
		/// </summary>
		public void Dispose()
		{
			if (channel_ != null)
			{
				channel_.StopListening(null);
				ChannelServices.UnregisterChannel(channel_);
				channel_ = null;
			}

			if (listener_ != null)
			{
				listener_.Stop();
				listener_.Close();
				listener_ = null;
			}

			StringBuilder sb = new StringBuilder();
			if (tuners_.Count > 0)
			{
				foreach (ITuner tuner in tuners_.Values)
				{
					tuner.Dispose();
				}
				tuners_.Clear();
				tunerFacades_.Clear();
				remoteTuners_ = null;

				sb.Append("Stream devices stopped and released.\r\n");
			}

			if (enableRecorder_ == true)
			{
				recorderManager_.Stop();

				sb.Append("Recorder stopped.\r\n");
			}

			if (sb.Length > 0)
			{
				Logger.Write(sb.ToString());
			}
		}

		/// <summary>
		/// T[rXJn܂B
		/// </summary>
		public void Start()
		{
			// Xg[|[gւ̎M񓯊ő҂
			listener_.Start();
			listener_.BeginGetContext(new AsyncCallback(AcceptEntry), null);

			// R[_[LȂJn
			if (enableRecorder_ == true)
			{
				recorderManager_.Start();
			}

			// [eBO̎MJn
			channel_.StartListening(null);
		}

		/// <summary>
		/// T[rX~܂B
		/// </summary>
		public void Stop()
		{
			// [eBO̎M~
			channel_.StopListening(null);

			// Xg[|[gւ̎M~
			listener_.Stop();

			// R[_[LȂ~
			if (enableRecorder_ == true)
			{
				recorderManager_.Stop();
			}
		}

		/// <summary>
		/// bZ[W𑗐M܂B
		/// </summary>
		/// <param name="response">X|XCX^X</param>
		/// <param name="statusCode">Xe[^XR[h</param>
		/// <param name="format">tH[}bg</param>
		/// <param name="args">ǉ</param>
		/// <remarks>O͔܂B</remarks>
		private void SendMessage(HttpListenerResponse response, HttpStatusCode statusCode,
			string format, params object[] args)
		{
			try
			{
				byte[] message = Encoding.UTF8.GetBytes(string.Format(format, args));

				response.StatusCode = (int)statusCode;
				response.ContentEncoding = Encoding.UTF8;
				response.ContentLength64 = message.Length;
				response.ContentType = "text/plain";
				response.KeepAlive = false;
				response.AddHeader("Server", this.GetType().Assembly.FullName);
				response.AddHeader("Pragma", "no-cache");
				response.AddHeader("Cache-Control", "no-cache, no-store, must-revalidate");

				using (Stream webStream = response.OutputStream)
				{
					webStream.Write(message, 0, message.Length);
					webStream.Close();
				}
			}
			catch
			{
			}
		}

		/// <summary>
		/// Xg[|[gւHTTPvMB
		/// </summary>
		/// <param name="ar">񓯊M҂v</param>
		private void AcceptEntry(IAsyncResult ar)
		{
			// v𑬂₩ɏÃXbh͉
			try
			{
				// MReLXg擾
				HttpListenerContext context = listener_.EndGetContext(ar);

				// ĎM񓯊ő҂
				listener_.BeginGetContext(new AsyncCallback(AcceptEntry), null);

				// MṽpX؂
				string absolutePath = context.Request.Url.AbsolutePath;

				// Sanity check
				if ((absolutePath.IndexOfAny(invalidPathChars_) == -1) &&
					(absolutePath.IndexOf(@"..") == -1) &&
					(absolutePath.IndexOf(@"//") == -1))
				{
					// [JEFuT[o[ւ̗vȂ
					string localRelativePath = absolutePath.Substring(streamBasePath_.Length).Replace('/', '\\');
					if (localRelativePath.StartsWith(@"\") == true)
					{
						localRelativePath = localRelativePath.Substring(1);
					}
					if (localRelativePath.Length == 0)
					{
						localRelativePath = "Default.dpr";
					}
					string localFilePath = Path.Combine(localWebServerRootPath_, localRelativePath);
					if (File.Exists(localFilePath) == true)
					{
						// Xbhv[ɓ
						ThreadPool.QueueUserWorkItem(new WaitCallback(LocalWebServerWorkItemEntry),
							new KeyValuePair<HttpListenerContext, string>(context, localFilePath));
						return;
					}
					else
					{
						// `[i[ւ̗vȂ
						// "/stream/CBonBlackTuner0.ts"
						ITuner tuner = null;
						if (tuners_.TryGetValue(absolutePath, out tuner) == true)
						{
							// Xg[~OsXbh𐶐
							// iԏs邽߁AXbhv[͎gȂj
							Thread thread = new Thread(new ParameterizedThreadStart(StreamingThreadEntry));
							thread.Start(new KeyValuePair<HttpListenerContext, ITuner>(context, tuner));
							return;
						}
					}
				}

				SendMessage(context.Response, HttpStatusCode.NotFound, "File \"{0}\" not found.", absolutePath);
				context.Response.Close();
			}
			catch (HttpListenerException)
			{
			}
			catch (ObjectDisposedException)
			{
			}
			catch (Exception ex)
			{
				Logger.Write(TraceType.Warning, ex);
			}
		}

		/// <summary>
		/// Xg[~OsXbh̃Gg|CgłB
		/// </summary>
		/// <param name="argument">HttpListenerContextƃ`[i[KeyValuePair</param>
		private void StreamingThreadEntry(object argument)
		{
			KeyValuePair<HttpListenerContext, ITuner> info =
				(KeyValuePair<HttpListenerContext, ITuner>)argument;

			string path = info.Key.Request.Url.AbsolutePath;
			IPEndPoint endPoint = info.Key.Request.RemoteEndPoint;

			Thread.CurrentThread.Name = string.Format("Stream sender: {0}", endPoint);

			Guid lockCookie = Guid.Empty;
			IDataQueue queue = null;
			try
			{
				try
				{
					string tuningSpaceString = info.Key.Request.QueryString["tuningSpace"];
					string channelString = info.Key.Request.QueryString["channel"];
					string programIDString = info.Key.Request.QueryString["programID"];
					string maxQueueSizeString = info.Key.Request.QueryString["maxQueueSize"];
					string lockedString = info.Key.Request.QueryString["locked"];

					StringBuilder sb = new StringBuilder();
					sb.AppendFormat("Stream connected: {0}", endPoint);

					// `[jOXy[Xƃ`lw肳Ă
					if ((string.IsNullOrEmpty(tuningSpaceString) == false) &&
						(string.IsNullOrEmpty(channelString) == false))
					{
						int tuningSpace = int.Parse(tuningSpaceString, CultureInfo.InvariantCulture);
						int channel = int.Parse(channelString, CultureInfo.InvariantCulture);

						// ύX݂
						if (string.IsNullOrEmpty(lockedString) == true)
						{
							info.Value.SetChannel(tuningSpace, channel, 0);
							sb.AppendFormat("\r\nTuning space: {0}-{1}", tuningSpace, channel);
						}
						else if (bool.Parse(lockedString) == false)
						{
							info.Value.SetChannel(tuningSpace, channel, 0);
							sb.AppendFormat("\r\nTuning space: {0}-{1}", tuningSpace, channel);
						}
						else
						{
							// bNvꍇ̓bN݂
							lockCookie = info.Value.SetChannelWithLock(tuningSpace, channel, 0);
							sb.AppendFormat("\r\nTuning space: {0}-{1} (Locked)", tuningSpace, channel);
						}
					}

					int programID = 0;
					int maxQueueSize = 100;

					// vOIDw肳ĂΎ擾
					if (string.IsNullOrEmpty(programIDString) == false)
					{
						programID = int.Parse(programIDString, CultureInfo.InvariantCulture);
						if (programID <= -1)
						{
							throw new ArgumentException("Invalid program ID.");
						}
						sb.AppendFormat("\r\nProgram ID: {0}", programID);
					}

					// L[TCYiMBjw肳ĂΎ擾
					if (string.IsNullOrEmpty(maxQueueSizeString) == false)
					{
						maxQueueSize = int.Parse(maxQueueSizeString, CultureInfo.InvariantCulture);
						if ((maxQueueSize <= 0) || (maxQueueSize > 1024))
						{
							throw new ArgumentException("Invalid max queue size.");
						}
						sb.AppendFormat("\r\nMax queue size: {0} MB", maxQueueSize);
					}

					// L[𐶐
					queue = info.Value.CreateQueue(programID, maxQueueSize * 1024 * 1024);

					Logger.Write(sb.ToString());
				}
				catch (Exception ex)
				{
					SendMessage(info.Key.Response, HttpStatusCode.InternalServerError,
						"{0}: {1}", ex.GetType().Name, ex.Message);
					throw;
				}

				info.Key.Response.StatusCode = (int)HttpStatusCode.OK;
				info.Key.Response.ContentType = streamContentType_;
				info.Key.Response.KeepAlive = false;

				info.Key.Response.AddHeader("Server", this.GetType().Assembly.FullName);
				info.Key.Response.AddHeader("Pragma", "no-cache");
				info.Key.Response.AddHeader("Cache-Control", "no-cache, no-store, must-revalidate");

				info.Key.Response.AddHeader("X-Interfaces", typeof(ITuner).Assembly.FullName);
				info.Key.Response.AddHeader("X-InteropServices", typeof(TunerFactory).Assembly.FullName);
				info.Key.Response.AddHeader("X-TunerDevice", info.Value.DeviceName);

				// o̓Xg[𓾂
				using (Stream webStream = info.Key.Response.OutputStream)
				{
					// Gh|Cgǉ
					lock (tunerFacades_)
					{
						tunerFacades_[path].AddEndPoint(endPoint, queue);
					}

					try
					{
						// Gh|Cgؒf邩AIRemoteTuner.DisconnectStreamĂяo鎖
						// Oă[v甲B
						while (true)
						{
							ITsData data = queue.Dequeue();
							webStream.Write(data.Data, 0, data.Size);
						}
					}
					finally
					{
						// Gh|Cg폜
						lock (tunerFacades_)
						{
							tunerFacades_[path].RemoveEndPoint(endPoint);
						}
					}
				}
			}
			catch (HttpListenerException)
			{
				Logger.Write("Connection closed: {0}", endPoint);
			}
			catch (ObjectDisposedException)
			{
				Logger.Write("Connection closed: {0}", endPoint);
			}
			catch (Exception ex)
			{
				Logger.Write(TraceType.Warning, ex);
			}
			finally
			{
				try
				{
					info.Key.Response.Close();
				}
				catch
				{
				}

				// L[蓖ĂĂΉ
				if (queue != null)
				{
					queue.Dispose();
				}

				// bNNbL[蓖ĂĂ΁AK
				if (lockCookie != Guid.Empty)
				{
					info.Value.Unlock(lockCookie);
				}
			}
		}

		/// <summary>
		/// DynamicPresentationRenderer̃CX^X𐶐Ȁ݂Ԃݒ肵܂B
		/// </summary>
		/// <returns>CX^X</returns>
		private DynamicPresentationRenderer CreateDynamicPresentationRenderer()
		{
			DynamicPresentationRenderer dpr = new DynamicPresentationRenderer("Information", "Node");

			// ݂̏ԂϐƂĐݒ肷
			EnvironmentValue environments = new EnvironmentValue("Environments");
			environments.Add(new EnvironmentValue("ServerName", externalAddresses_[0]));
			environments.Add(new EnvironmentValue("RemoteControlUrl", remoteControlUrl_.ToString()));
			environments.Add(new EnvironmentValue("OsVersion", Environment.OSVersion.VersionString));
			environments.Add(new EnvironmentValue("ProcessorCount", Environment.ProcessorCount.ToString(CultureInfo.InvariantCulture)));
			environments.Add(new EnvironmentValue("WorkingSet", Environment.WorkingSet.ToString(CultureInfo.InvariantCulture)));
			environments.Add(new EnvironmentValue("NetCLR", Environment.Version.ToString()));
			environments.Add(new EnvironmentValue("mscorlib", typeof(object).Assembly.FullName));
			environments.Add(new EnvironmentValue("Interfaces", typeof(ITuner).Assembly.FullName));
			environments.Add(new EnvironmentValue("InteropServices", typeof(TunerFactory).Assembly.FullName));
			environments.Add(new EnvironmentValue("StreamService", this.GetType().Assembly.FullName));
			environments.Add(new EnvironmentValue("DateTime", DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss")));
			dpr.Add(environments);

			// `[i[Q
			EnvironmentValue tuners = new EnvironmentValue("Tuners");
			foreach (KeyValuePair<string, ITuner> entry in tuners_)
			{
				EnvironmentValue tuner = new EnvironmentValue("Tuner");
				tuner.Add(new EnvironmentValue("TunerName", entry.Value.TunerName));
				tuner.Add(new EnvironmentValue("DeviceID", entry.Value.DeviceID.ToString(CultureInfo.InvariantCulture)));
				tuner.Add(new EnvironmentValue("DeviceName", entry.Value.DeviceName));
				tuner.Add(new EnvironmentValue("EmmProcess", entry.Value.EnableEmmProcess.ToString()));
				tuner.Add(new EnvironmentValue("LnbPower", entry.Value.EnableLnbPower.ToString()));
				tuner.Add(new EnvironmentValue("SignalLevel", entry.Value.SignalLevel.ToString(CultureInfo.InvariantCulture)));
				tuner.Add(new EnvironmentValue("TuningSpace", entry.Value.TuningSpace.ToString(CultureInfo.InvariantCulture)));
				tuner.Add(new EnvironmentValue("Channel", entry.Value.Channel.ToString(CultureInfo.InvariantCulture)));
				tuner.Add(new EnvironmentValue("IsLocked", entry.Value.IsLocked.ToString(CultureInfo.InvariantCulture)));
				tuner.Add(new EnvironmentValue("StreamUrl", entry.Key));
				tuners.Add(tuner);
			}
			dpr.Add(tuners);

			// R[_[
			EnvironmentValue recorder = new EnvironmentValue("Recorder");
			recorder.Add(new EnvironmentValue("Enable", enableRecorder_.ToString()));
			EnvironmentValue storages = new EnvironmentValue("Storages");
			foreach (string storageID in recorderManager_.StorageIDs)
			{
				EnvironmentValue storage = new EnvironmentValue("Storage");
				storage.Add(new EnvironmentValue("StorageID", storageID));
				storage.Add(new EnvironmentValue("FreeSpace", (recorderManager_.GetCurrentSpaces(storageID) / 1024 / 1024).ToString(CultureInfo.InvariantCulture)));
				storages.Add(storage);
			}
			recorder.Add(storages);
			EnvironmentValue entries = new EnvironmentValue("Entries");
			foreach (IRegisteredProgram program in recorderManager_)
			{
				EnvironmentValue entry = new EnvironmentValue("Entry");
				entry.Add(new EnvironmentValue("RegisteredID", program.RegisteredID.ToString()));
				entry.Add(new EnvironmentValue("RegisteredTime", program.RegisteredTime.ToString("yyyy/MM/dd HH:mm:ss")));
				entry.Add(new EnvironmentValue("FileName", program.FileName));
				entry.Add(new EnvironmentValue("StartTime", program.StartTime.ToString("yyyy/MM/dd HH:mm:ss")));
				entry.Add(new EnvironmentValue("Duration", program.Duration.ToString()));
				entry.Add(new EnvironmentValue("TuningSpace", program.TuningSpace.ToString(CultureInfo.InvariantCulture)));
				entry.Add(new EnvironmentValue("Channel", program.Channel.ToString(CultureInfo.InvariantCulture)));
				entry.Add(new EnvironmentValue("ProgramID", program.ProgramID.ToString(CultureInfo.InvariantCulture)));
				entry.Add(new EnvironmentValue("TunerStreamUrl", program.TunerStreamUrl.ToString()));
				entry.Add(new EnvironmentValue("StorageID", program.StorageID));
				entry.Add(new EnvironmentValue("RecordingStatus", recorderManager_.GetRecordingStatus(program.RegisteredID).ToString()));
				entries.Add(entry);
			}
			recorder.Add(entries);
			dpr.Add(recorder);

			return dpr;
		}

		/// <summary>
		/// [JEFuT[o[sXbhv[̃Gg|CgłB
		/// </summary>
		/// <param name="argument">HttpListenerContext</param>
		private void LocalWebServerWorkItemEntry(object argument)
		{
			KeyValuePair<HttpListenerContext, string> info = (KeyValuePair<HttpListenerContext, string>)argument;

			string absolutePath = info.Key.Request.Url.AbsolutePath;
			string filePath = info.Value;
			IPEndPoint endPoint = info.Key.Request.RemoteEndPoint;

			Thread.CurrentThread.Name = string.Format("Local web server: {0}", endPoint);

			Stream contentStream = null;
			try
			{
				try
				{
					// t@CI[v
					contentStream = new FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read);
				}
				catch (Exception ex)
				{
					SendMessage(info.Key.Response, HttpStatusCode.InternalServerError,
						"{0}: {1}", ex.GetType().Name, ex.Message);
					throw;
				}

				info.Key.Response.StatusCode = (int)HttpStatusCode.OK;
				info.Key.Response.KeepAlive = false;

				info.Key.Response.AddHeader("Server", this.GetType().Assembly.FullName);
				info.Key.Response.AddHeader("Pragma", "no-cache");
				info.Key.Response.AddHeader("Cache-Control", "no-cache, no-store, must-revalidate");

				string extension = Path.GetExtension(filePath);
				string contentType = ContentTypeMapper.Map(extension);
				if (string.IsNullOrEmpty(contentType) == false)
				{
					info.Key.Response.ContentType = contentType;
				}
				else
				{
					info.Key.Response.ContentType = ContentTypeMapper.OctetStream;
					info.Key.Response.AddHeader("Content-Disposition",
						string.Format("attachment; filename=\"{0}\"", Path.GetFileName(filePath)));
				}

				Logger.Write("Send local file: {0}\r\nFile path: {1}\r\nFile length: {2}",
					endPoint, filePath, contentStream.Length);

				// o̓Xg[𓾂
				using (Stream webStream = info.Key.Response.OutputStream)
				{
					// DPR`
					if (string.Compare(extension, ".dpr", true) == 0)
					{
						// DPR_O
						DynamicPresentationRenderer dpr = CreateDynamicPresentationRenderer();
						dpr.Render(contentStream, webStream);
					}
					else
					{
						// t@C𑗐M
						byte[] buffer = new byte[1 * 1024 * 1024];
						while (true)
						{
							int read = contentStream.Read(buffer, 0, buffer.Length);
							if (read == 0)
							{
								break;
							}
							webStream.Write(buffer, 0, read);
						}
					}
					webStream.Close();
				}
			}
			catch (HttpListenerException)
			{
				Logger.Write("Connection closed: {0}", endPoint);
			}
			catch (ObjectDisposedException)
			{
				Logger.Write("Connection closed: {0}", endPoint);
			}
			catch (Exception ex)
			{
				Logger.Write(TraceType.Warning, ex);
			}
			finally
			{
				try
				{
					info.Key.Response.Close();
				}
				catch
				{
				}
				if (contentStream != null)
				{
					contentStream.Close();
				}
			}
		}
	}
}
