//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Collections;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Drawing.Design;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.ServiceProcess;
using System.Text;
using Microsoft.Win32;

namespace Bon
{
	/// <summary>
	/// Xg[T[rX̃WXg`𒊏ۉNXłB
	/// </summary>
	[TypeConverter(typeof(ExpandableObjectConverter))]
	internal sealed class StreamServiceRegistryAccessor : IDisposable
	{
		#region Interop
		[DllImport("shell32.dll")]
		private static extern bool SHGetSpecialFolderPath(IntPtr hwndOwner,
		   [Out] StringBuilder lpszPath, int nFolder, bool fCreate);

		// All Users\My Video
		private static readonly int CSIDL_COMMON_VIDEO = 0x0037;
		#endregion

		private readonly bool writable_;
		private readonly CultureInfo cultureInfo_;
		private RegistryKey key_;
		
		/// <summary>
		/// RXgN^łB
		/// </summary>
		/// <param name="writable">݂\ɂꍇtrue</param>
		/// <param name="cultureInfo">l̎擾EݒɎgpJ`</param>
		public StreamServiceRegistryAccessor(bool writable, CultureInfo cultureInfo)
		{
			writable_ = writable;
			cultureInfo_ = cultureInfo;
			key_ = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\BonSDK\" + StreamService.ServiceName, writable_);
		}

		/// <summary>
		/// t@CiCUłB
		/// </summary>
		~StreamServiceRegistryAccessor()
		{
			Dispose();
		}

		/// <summary>
		/// Dispose\bhłB
		/// </summary>
		public void Dispose()
		{
			if (key_ != null)
			{
				key_.Close();
				key_ = null;
				GC.SuppressFinalize(this);
			}
		}

		/// <summary>
		/// T[rXN邩ǂ擾Eݒ肵܂B
		/// </summary>
		[Category("General")]
		[DefaultValue(true)]
		[Description("Auto start stream service.")]
		public bool AutoStart
		{
			get
			{
				using (RegistryKey key = Registry.LocalMachine.OpenSubKey(
					@"SYSTEM\CurrentControlSet\Services\" + StreamService.ServiceName, false))
				{
					int start = Convert.ToInt32(key.GetValue("Start", 3), CultureInfo.InvariantCulture);
					if (start == 2)
					{
						return true;
					}
					else
					{
						return false;
					}
				}
			}
			set
			{
				using (RegistryKey key = Registry.LocalMachine.OpenSubKey(
					@"SYSTEM\CurrentControlSet\Services\" + StreamService.ServiceName, true))
				{
					key.SetValue("Start", value ? 2 : 3);
				}
			}
		}

		/// <summary>
		/// [JEFuT[oLɂ邩ǂ擾Eݒ肵܂B
		/// </summary>
		[Category("General")]
		[DefaultValue(true)]
		[Description("Enable local web server.")]
		public bool EnableLocalWebServer
		{
			get
			{
				return Convert.ToInt32(key_.GetValue("EnableLocalWebServer", 1), cultureInfo_) > 0;
			}
			set
			{
				key_.SetValue("EnableLocalWebServer", value ? 1 : 0);
			}
		}

		/// <summary>
		/// ftHg̃[JEFuT[ozMtH_ւ̃pX擾܂B
		/// </summary>
		/// <returns></returns>
		public static string GetDefaultLocalWebServerRootPath()
		{
			Assembly assembly = Assembly.GetExecutingAssembly();
			return Path.Combine(Path.GetDirectoryName(assembly.Location), "wwwroot");
		}

		/// <summary>
		/// [JEFuT[ozMtH_̃pX擾Eݒ肵܂B
		/// </summary>
		[Category("General")]
		[Description("Local web server root path. (Root path mapped to StreamBasePath)")]
		public string LocalWebServerRootPath
		{
			get
			{
				return Convert.ToString(key_.GetValue(
					"LocalWebServerRootPath", GetDefaultLocalWebServerRootPath()), cultureInfo_);
			}
			set
			{
				if (string.IsNullOrEmpty(value) == true)
				{
					throw new ArgumentException("Invalid setup file path.");
				}
				key_.SetValue("LocalWebServerRootPath", value);
			}
		}

		private bool ShouldSerializeLocalWebServerRootPath()
		{
			return LocalWebServerRootPath != GetDefaultLocalWebServerRootPath();
		}

		/// <summary>
		/// ftHg̃[JEFuT[ozMtH_̃pXݒ肵܂B
		/// </summary>
		public void ResetLocalWebServerRootPath()
		{
			LocalWebServerRootPath = GetDefaultLocalWebServerRootPath();
		}

		/// <summary>
		/// Xg[~OɎgp|[gԍ擾Eݒ肵܂B
		/// </summary>
		[Category("Streaming")]
		[DefaultValue(26110)]
		[Description("Using HTTP streaming port.")]
		public int StreamPort
		{
			get
			{
				return Convert.ToInt32(key_.GetValue("StreamPort", 26110), cultureInfo_);
			}
			set
			{
				if (value <= 0)
				{
					throw new ArgumentException("Invalid stream port.");
				}
				key_.SetValue("StreamPort", value);
			}
		}

		/// <summary>
		/// Xg[~OɎgp郊\[Xx[X擾Eݒ肵܂B
		/// </summary>
		[Category("Streaming")]
		[DefaultValue("/stream")]
		[Description("Using HTTP resource base.")]
		public string StreamBasePath
		{
			get
			{
				return Convert.ToString(key_.GetValue("StreamBasePath", "/stream"), cultureInfo_);
			}
			set
			{
				if (string.IsNullOrEmpty(value) == true)
				{
					throw new ArgumentException("Invalid stream base path.");
				}
				key_.SetValue("StreamBasePath", value);
			}
		}

		/// <summary>
		/// Xg[~OɎgptH[}bg\ȃ\[X擾Eݒ肵܂B
		/// </summary>
		[Category("Streaming")]
		[DefaultValue("{0}{1}.ts")]
		[Description("Using HTTP resource name. \"{0}\" is tuner class name, \"{1}\" is tuner ID number.")]
		public string StreamNameFormat
		{
			get
			{
				return Convert.ToString(key_.GetValue("StreamNameFormat", "{0}{1}.ts"), cultureInfo_);
			}
			set
			{
				if (string.IsNullOrEmpty(value) == true)
				{
					throw new ArgumentException("Invalid stream name format.");
				}
				string.Format(value, "Test", 0);
				key_.SetValue("StreamNameFormat", value);
			}
		}

		/// <summary>
		/// Xg[~OɎgpContentType擾Eݒ肵܂B
		/// </summary>
		[Category("Streaming")]
		[DefaultValue("video/mpeg")]
		[Description("Using HTTP streaming content type.")]
		public string StreamContentType
		{
			get
			{
				return Convert.ToString(key_.GetValue("StreamContentType", "video/mpeg"), cultureInfo_);
			}
			set
			{
				if (string.IsNullOrEmpty(value) == true)
				{
					throw new ArgumentException("Invalid content type.");
				}
				key_.SetValue("StreamContentType", value);
			}
		}

		/// <summary>
		/// [eBOɎgp|[gԍ擾Eݒ肵܂B
		/// </summary>
		[Category("Remoting")]
		[DefaultValue(26111)]
		[Description("Using .NET remoting port.")]
		public int RemotePort
		{
			get
			{
				return Convert.ToInt32(key_.GetValue("RemotePort", 26111), cultureInfo_);
			}
			set
			{
				if (value <= 0)
				{
					throw new ArgumentException("Invalid remote port.");
				}
				key_.SetValue("RemotePort", value);
			}
		}

		/// <summary>
		/// Xg[~OɎgpURIx[X擾Eݒ肵܂B
		/// </summary>
		[Category("Remoting")]
		[DefaultValue("/remote")]
		[Description("Using .NET URI base.")]
		public string RemoteBasePath
		{
			get
			{
				return Convert.ToString(key_.GetValue("RemoteBasePath", "/remote"), cultureInfo_);
			}
			set
			{
				if (string.IsNullOrEmpty(value) == true)
				{
					throw new ArgumentException("Invalid remote base path.");
				}
				key_.SetValue("RemoteBasePath", value);
			}
		}

		/// <summary>
		/// R[_@\̗L擾Eݒ肵܂B
		/// </summary>
		[Category("Recorder")]
		[DefaultValue(true)]
		[Description("Enable recorder service.")]
		public bool EnableRecorder
		{
			get
			{
				return Convert.ToInt32(key_.GetValue("EnableRecorder", 1), cultureInfo_) > 0;
			}
			set
			{
				key_.SetValue("EnableRecorder", value ? 1 : 0);
			}
		}

		/// <summary>
		/// R[_̘^Xg[WŒeʂ擾Eݒ肵܂B
		/// </summary>
		[Category("Recorder")]
		[DefaultValue((long)(100 * 1024 * 1024))]
		[Description("Allow recording if storage spaces is larger than this value.")]
		public long MinimumSpaceRequirements
		{
			get
			{
				return Convert.ToInt64(key_.GetValue("MinimumSpaceRequirements", 100), cultureInfo_) * 1024 * 1024;
			}
			set
			{
				int rawValue = (int)(value / 1024 / 1024);
				if (rawValue <= 0)
				{
					throw new ArgumentException("Invalid storage space size.");
				}
				key_.SetValue("MinimumSpaceRequirements", rawValue);
			}
		}

		/// <summary>
		/// R[_^悷ۂɓ]Ɏgpobt@TCY擾Eݒ肵܂B
		/// </summary>
		[Category("Recorder")]
		[DefaultValue(4 * 1024 * 1024)]
		[Description("Using buffer size with recording.")]
		public int BufferSize
		{
			get
			{
				return Convert.ToInt32(key_.GetValue("BufferSize", 4), cultureInfo_) * 1024 * 1024;
			}
			set
			{
				int rawValue = value / 1024 / 1024;
				if (rawValue <= 0)
				{
					throw new ArgumentException("Invalid buffer size.");
				}
				key_.SetValue("BufferSize", rawValue);
			}
		}

		/// <summary>
		/// ftHg̗\^̎OE̎ԕ擾܂B
		/// </summary>
		/// <returns>\^̎OE̎ԕ</returns>
		public static TimeSpan GetDefaultExtraTimeSpan()
		{
			return new TimeSpan(0, 0, 5);
		}

		/// <summary>
		/// \^̎OE̎ԕ擾Eݒ肵܂B
		/// </summary>
		[Category("Recorder")]
		[Description("Extra time span with registered recording.")]
		public TimeSpan ExtraTimeSpan
		{
			get
			{
				return TimeSpan.Parse(Convert.ToString(
					key_.GetValue("ExtraTimeSpan", GetDefaultExtraTimeSpan().ToString()), cultureInfo_));
			}
			set
			{
				key_.SetValue("ExtraTimeSpan", value.ToString());
			}
		}

		private bool ShouldSerializeExtraTimeSpan()
		{
			return ExtraTimeSpan != GetDefaultExtraTimeSpan();
		}

		/// <summary>
		/// \^̎OE̎ԕftHgɐݒ肵܂B
		/// </summary>
		public void ResetExtraTimeSpan()
		{
			ExtraTimeSpan = GetDefaultExtraTimeSpan();
		}

		/// <summary>
		/// ftHg̃f[^x[Xt@Cւ̃pX擾܂B
		/// </summary>
		/// <returns>f[^x[Xt@Cւ̃pX</returns>
		public static string GetDefaultDatabasePath()
		{
			string folderPath = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
			return Path.Combine(folderPath, @"BonSDK\Recorder.db");
		}

		/// <summary>
		/// R[_\ێf[^x[Xt@Cւ̃pX擾Eݒ肵܂B
		/// </summary>
		[Category("Recorder")]
		[Description("Recorder states database path.")]
		public string DatabasePath
		{
			get
			{
				return Convert.ToString(key_.GetValue("DatabasePath", GetDefaultDatabasePath()), cultureInfo_);
			}
			set
			{
				if (string.IsNullOrEmpty(value) == true)
				{
					throw new ArgumentException("Invalid database path.");
				}
				key_.SetValue("DatabasePath", value);
			}
		}

		private bool ShouldSerializeDatabasePath()
		{
			return DatabasePath != GetDefaultDatabasePath();
		}

		/// <summary>
		/// R[_\ێf[^x[Xt@Cւ̃pXftHgɐݒ肵܂B
		/// </summary>
		public void ResetDatabasePath()
		{
			DatabasePath = GetDefaultDatabasePath();
		}

		/// <summary>
		/// w肳ꂽRNVmr܂B
		/// </summary>
		/// <param name="lhs">r</param>
		/// <param name="rhs">rΏ</param>
		/// <returns>evĂtrue</returns>
		private static bool DictionaryEquals(IDictionary<string, string> lhs, IDictionary<string, string> rhs)
		{
			if (lhs.Count != rhs.Count)
			{
				return false;
			}

			foreach (KeyValuePair<string, string> lvalue in lhs)
			{
				string value = null;
				if (rhs.TryGetValue(lvalue.Key, out value) == false)
				{
					return true;
				}
				if (lvalue.Value != value)
				{
					return false;
				}
			}

			return true;
		}

		/// <summary>
		/// ftHg̃Xg[Wւ̃tH[}bg\ȃpXQ擾܂B
		/// </summary>
		/// <returns>Xg[Wւ̃tH[}bg\ȃpXQ</returns>
		public static IDictionary<string, string> GetDefaultStoragePaths()
		{
			SortedDictionary<string, string> storagePaths = new SortedDictionary<string, string>();
			StringBuilder folderPath = new StringBuilder(260);
			if (SHGetSpecialFolderPath(IntPtr.Zero, folderPath, CSIDL_COMMON_VIDEO, true) == true)
			{
				storagePaths["Default"] = Path.Combine(folderPath.ToString(),
					@"Bon.StreamService\{8}_{9}_{10}\{0}_{2:yyyyMMdd}_{2:HHmmss}_{5:yyyyMMdd}_{5:HHmmss}.ts");
			}
			else
			{
				storagePaths["Default"] = Path.Combine(
					Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData),
					@"BonSDK\Storage\{8}_{9}_{10}\{0}_{2:yyyyMMdd}_{2:HHmmss}_{5:yyyyMMdd}_{5:HHmmss}.ts");
			}
			return storagePaths;
		}

		/// <summary>
		/// Xg[Wւ̃tH[}bg\ȃpXQ擾܂B
		/// </summary>
		[Category("Recorder")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
		[Editor(typeof(SelectableDictionaryEditor<StreamServiceRegistryAccessor.StoragePath>), typeof(UITypeEditor))]
		[Description("Definitions of storage paths. (Name is identity of storage, Value is formattable path)")]
		public RegistryDictionary StoragePaths
		{
			get
			{
				return new RegistryDictionary(key_, "StoragePaths", writable_, cultureInfo_);
			}
		}

		private bool ShouldSerializeStoragePaths()
		{
			using (RegistryDictionary storagePaths = StoragePaths)
			{
				return DictionaryEquals(storagePaths, GetDefaultStoragePaths()) == false;
			}
		}

		/// <summary>
		/// Xg[Wւ̃tH[}bg\ȃpXQftHgɐݒ肵܂B
		/// </summary>
		public void ResetStoragePaths()
		{
			using (RegistryDictionary storagePaths = StoragePaths)
			{
				storagePaths.Clear();
				foreach (KeyValuePair<string, string> entry in GetDefaultStoragePaths())
				{
					storagePaths.Add(entry.Key, entry.Value);
				}
			}
		}

		/// <summary>
		/// ftHgŎgpv[[\tgւ̃ftHg̃pXQ擾܂B
		/// </summary>
		/// <returns>v[[\tgւ̃pXQ</returns>
		public static IDictionary<string, string> GetDefaultDefaultPlayers()
		{
			SortedDictionary<string, string> defaultPlayers = new SortedDictionary<string, string>();
			defaultPlayers["Windows-INTEL"] = @"%ProgramFiles%\VideoLAN\VLC\vlc.exe";
			defaultPlayers["Windows-AMD64"] = @"%ProgramFiles(x86)%\VideoLAN\VLC\vlc.exe";
			return defaultPlayers;
		}

		/// <summary>
		/// ftHgŎgpv[[\tgւ̃pXQ擾܂B
		/// </summary>
		[Category("Player")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
		[Editor(typeof(SelectableDictionaryEditor<StreamServiceRegistryAccessor.DefaultPlayer>), typeof(UITypeEditor))]
		[Description("Definitions of default player application paths. (Name is identity of platform, Value is application path)")]
		public RegistryDictionary DefaultPlayers
		{
			get
			{
				return new RegistryDictionary(key_, "DefaultPlayers", writable_, cultureInfo_);
			}
		}

		private bool ShouldSerializeDefaultPlayers()
		{
			using (RegistryDictionary defaultPlayers = DefaultPlayers)
			{
				return DictionaryEquals(defaultPlayers, GetDefaultDefaultPlayers()) == false;
			}
		}

		/// <summary>
		/// ftHgŎgpv[[\tgւ̃pXQftHgɐݒ肵܂B
		/// </summary>
		public void ResetDefaultPlayers()
		{
			using (RegistryDictionary defaultPlayers = DefaultPlayers)
			{
				defaultPlayers.Clear();
				foreach (KeyValuePair<string, string> entry in GetDefaultDefaultPlayers())
				{
					defaultPlayers.Add(entry.Key, entry.Value);
				}
			}
		}

		/// <summary>
		/// Xg[Wւ̃tH[}bg\ȃftHgpX`NXłB
		/// </summary>
		public sealed class StoragePath : SelectableNameValueBase
		{
			/// <summary>
			/// RXgN^łB
			/// </summary>
			public StoragePath()
			{
			}

			/// <summary>
			/// l擾Eݒ肵܂B
			/// </summary>
			[TypeConverter(typeof(SelectableValueConverter<StoragePath.DefaultNames>))]
			[Description("Storage ID")]
			public override string Name
			{
				get
				{
					return base.Name;
				}
				set
				{
					base.Name = value;
				}
			}

			/// <summary>
			/// l擾Eݒ肵܂B
			/// </summary>
			[TypeConverter(typeof(SelectableValueConverter<StoragePath.DefaultValues>))]
			[Description("Formattable storage path. ({0}:FileName, {1}:RegisteredID, {2}:StartTime, {3}:StartTime(UTC), {4}:Duration, {5}:RealStartTime, {6}:RealStartTime(UTC), {7}:RealDuration, {8}:TuningSpace, {9}:Channel, {10}:ProgramID, {11}:RegisteredTime, {12}:RegisteredTime(UTC))")]
			public override string Value
			{
				get
				{
					return base.Value;
				}
				set
				{
					base.Value = value;
				}
			}

			public sealed class DefaultNames : List<string>
			{
				public DefaultNames()
				{
					IDictionary<string, string> storagePaths = StreamServiceRegistryAccessor.GetDefaultStoragePaths();
					base.AddRange(storagePaths.Keys);
				}
			}

			public sealed class DefaultValues : List<string>
			{
				public DefaultValues()
				{
					IDictionary<string, string> storagePaths = StreamServiceRegistryAccessor.GetDefaultStoragePaths();
					base.AddRange(storagePaths.Values);
				}
			}
		}

		/// <summary>
		/// ftHgŎgpv[[\tg̃ftHgpX`NXłB
		/// </summary>
		public sealed class DefaultPlayer : SelectableNameValueBase
		{
			/// <summary>
			/// RXgN^łB
			/// </summary>
			public DefaultPlayer()
			{
			}

			/// <summary>
			/// l擾Eݒ肵܂B
			/// </summary>
			[TypeConverter(typeof(SelectableValueConverter<DefaultPlayer.DefaultNames>))]
			[Description("Platform ID. ([PlatformName]-[ArchitectureName], Ex:\"Windows-INTEL\",\"Windows-AMD64\")")]
			public override string Name
			{
				get
				{
					return base.Name;
				}
				set
				{
					base.Name = value;
				}
			}

			/// <summary>
			/// l擾Eݒ肵܂B
			/// </summary>
			[TypeConverter(typeof(SelectableValueConverter<DefaultPlayer.DefaultValues>))]
			[Description("Application path.")]
			public override string Value
			{
				get
				{
					return base.Value;
				}
				set
				{
					base.Value = value;
				}
			}

			public sealed class DefaultNames : List<string>
			{
				public DefaultNames()
				{
					IDictionary<string, string> storagePaths = StreamServiceRegistryAccessor.GetDefaultDefaultPlayers();
					base.AddRange(storagePaths.Keys);
				}
			}

			public sealed class DefaultValues : List<string>
			{
				public DefaultValues()
				{
					IDictionary<string, string> storagePaths = StreamServiceRegistryAccessor.GetDefaultDefaultPlayers();
					base.AddRange(storagePaths.Values);
				}
			}
		}
	}
}
