//////////////////////////////////////////////////////////////////////////
// BonSDK library header file
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// SmartThread.h: X}[gXbhNX
//
/////////////////////////////////////////////////////////////////////////////

#pragma once


namespace Bon
{

/////////////////////////////////////////////////////////////////////////////
// VOiC^[tFCX
/////////////////////////////////////////////////////////////////////////////

class ISignalInstance : public ISignal
{
public:
	virtual void Release() = 0;
};

/////////////////////////////////////////////////////////////////////////////
// tOVOiNX
/////////////////////////////////////////////////////////////////////////////

class BONAPI CFlagSignal : public ISignalInstance
{
public:
	CFlagSignal(DWORD dwResolution = 100);

	const IKernelObject *GetKernelObject() const;

	const bool Set();
	const bool Reset();

	operator bool() const;
	bool operator !() const;

	const WaitResult Wait(const DWORD dwTimeout = INFINITE, const bool pumpMessage = true) const;

	void Release();

private:
	const DWORD m_dwResolution;
	bool m_bFlag;
};


/////////////////////////////////////////////////////////////////////////////
// CxgVOiNX
/////////////////////////////////////////////////////////////////////////////

class BONAPI CEventSignal : public ISignalInstance
{
public:
	CEventSignal();

	const IKernelObject *GetKernelObject() const;

	const bool Set();
	const bool Reset();

	operator bool() const;
	bool operator !() const;

	const WaitResult Wait(const DWORD dwTimeout = INFINITE, const bool pumpMessage = true) const;

	void Release();

private:
	CSmartEvent m_Event;
};


/////////////////////////////////////////////////////////////////////////////
// X}[gXbhNX
/////////////////////////////////////////////////////////////////////////////

class BONAPI CSmartThreadBase : public CKernelObject
{
public:
	~CSmartThreadBase(void);

	void SetThreadName(LPCTSTR pThreadName);
	const bool SetPriority(int nPriority);
	const bool EndThread(const bool pumpMessage = false);

	const bool SuspendThread(void);
	const bool ResumeThread(void);

	const DWORD GetThreadID(void) const;
	const bool IsRunning(void) const;
	
protected:
	CSmartThreadBase(ISignalInstance *pKillSignal);

	const bool StartThread(LPCTSTR pThreadName, const ULONG_PTR threadParam, const bool bStart);

	virtual void ThreadEntry(const ISignal &killSignal, const ULONG_PTR threadParam) = 0;

private:
	static unsigned int WINAPI ThreadFunc(void *pParam);

	ULONG_PTR m_ThreadParam;
	CBonString m_ThreadName;
	DWORD m_dwThreadID;
	CSmartEvent m_SyncSignal;
	ISignalInstance *m_pKillSignal;
};


/////////////////////////////////////////////////////////////////////////////
// X}[gXbhNX
/////////////////////////////////////////////////////////////////////////////

template <class T> class CSmartThread : public CSmartThreadBase
{
public:
	// Xbh\bh^
	typedef void (T::*TAGET_FUNC)(CSmartThread<T> &thread, const ISignal &killSignal, const ULONG_PTR param);

	CSmartThread(ISignalInstance *pKillSignal = new CEventSignal())
		: CSmartThreadBase(pKillSignal)
		, m_pThreadOwner(NULL)
		, m_pTargetFunc(NULL)
	{
	}

	const bool StartThread(LPCTSTR pThreadName, T *pOwner, TAGET_FUNC pTargetFunc, const ULONG_PTR threadParam = 0, const bool bStart = true)
	{
		if(!pOwner || !pTargetFunc)return false;
		
		// ֐|C^Ap[^ۑ
		m_pThreadOwner = pOwner;
		m_pTargetFunc = pTargetFunc;
		
		return CSmartThreadBase::StartThread(pThreadName, threadParam, bStart);
	}
	
protected:
	void ThreadEntry(const ISignal &killSignal, const ULONG_PTR threadParam)
	{
		// Xbh\bhs
		(m_pThreadOwner->*m_pTargetFunc)(*this, killSignal, threadParam);
	}

private:
	T *m_pThreadOwner;
	TAGET_FUNC m_pTargetFunc;
};

}
