//////////////////////////////////////////////////////////////////////////
// BonSDK CAS card reader library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// CasCardReader.cpp: CASJ[h[_NX
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "BonSDK.h"
#include <Memory>
#include <TChar.h>
#include "CasCardReader.h"


#pragma comment(lib, "WinScard.lib")


/////////////////////////////////////////////////////////////////////////////
// t@C[J萔ݒ
/////////////////////////////////////////////////////////////////////////////

// foCX
#define DEVICENAME	TEXT("CAS ėpX}[gJ[h[_")


/////////////////////////////////////////////////////////////////////////////
// X}[gJ[h[_@gUNVێNX
/////////////////////////////////////////////////////////////////////////////

class SCardTransaction
{
private:
	SCARDHANDLE handle_;

public:
	SCardTransaction(SCARDHANDLE handle)
		: handle_(handle)
	{
		::SCardBeginTransaction(handle_);
	}

	~SCardTransaction()
	{
		::SCardEndTransaction(handle_, SCARD_LEAVE_CARD);
	}
};


/////////////////////////////////////////////////////////////////////////////
// ėpX}[gJ[h[_p CASJ[hC^tF[XNX
/////////////////////////////////////////////////////////////////////////////

const BONGUID CCasCardReader::GetDeviceType(void) const
{
	// foCX̃^CvԂ
	return ::BON_NAME_TO_GUID(TEXT("IHalCasCard"));
}

const CBonString CCasCardReader::GetDeviceName() const
{
	static const CBonString deviceName(DEVICENAME);

	// foCXԂ
	return deviceName;
}

const DWORD CCasCardReader::GetTotalDeviceNum(void) const
{
	// CASJ[h}ĂJ[h[_Ԃ

	// X}[gJ[h[_
	if(!EnumCasCardReader())return 0UL;

	SCARDHANDLE hCasCard;
	DWORD dwCasCardNum = 0UL;
	DWORD dwActiveProtocol;
	
	// s
	for(DWORD dwIndex = 0UL ; dwIndex < m_CardReaderArray.size() ; dwIndex++){

		hCasCard = NULL;
		dwActiveProtocol = SCARD_PROTOCOL_UNDEFINED;

		// J[h[_ɐڑ
		if(::SCardConnect(m_ScardContext, m_CardReaderArray[dwIndex], SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1, &hCasCard, &dwActiveProtocol) == SCARD_S_SUCCESS){
			// vgR`FbN
			if(dwActiveProtocol == SCARD_PROTOCOL_T1){
				// R}hM
				if(InitialSetting(hCasCard)){
					// CASJ[hCNg
					dwCasCardNum++;
					}
				}
			// J[h[_ؒf
			::SCardDisconnect(hCasCard, SCARD_LEAVE_CARD);	
			}
		}

	return dwCasCardNum;
}

const DWORD CCasCardReader::GetActiveDeviceNum(void) const
{
	// X}[gJ[h[_͋p\Ȃߏ0Ԃ
	return 0UL;
}

const bool CCasCardReader::OpenCard(void)
{
	// UN[Y
	CloseCard();

	// X}[gJ[h[_
	if(!EnumCasCardReader())return false;

	DWORD dwActiveProtocol;

	// s
	for(DWORD dwIndex = 0UL ; dwIndex < m_CardReaderArray.size() ; dwIndex++){

		m_hCasCard = NULL;
		dwActiveProtocol = SCARD_PROTOCOL_UNDEFINED;

		// J[h[_ɐڑ
		if(::SCardConnect(m_ScardContext, m_CardReaderArray[dwIndex], SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1, &m_hCasCard, &dwActiveProtocol) == SCARD_S_SUCCESS){
			// vgR`FbN
			if(dwActiveProtocol == SCARD_PROTOCOL_T1){
				// R}hM
				if(InitialSetting(m_hCasCard)){
					// CASJ[hCNg
					break;
					}
				}

			// J[h[_ؒf
			::SCardDisconnect(m_hCasCard, SCARD_LEAVE_CARD);
			m_hCasCard = NULL;
			}
		}

	return (m_hCasCard)? true : false;
}

void CCasCardReader::CloseCard(void)
{
	// J[h[_ؒf
	if(m_hCasCard){
		::SCardDisconnect(m_hCasCard, SCARD_LEAVE_CARD);
		m_hCasCard = NULL;
		}
}

const bool CCasCardReader::IsInteractive(void)
{
	// oʐM̉ۂԂ
	return true;
}

const BYTE * CCasCardReader::GetCardID(void)
{
	// Card ID Ԃ
	return (m_hCasCard)? m_CasCardInfo.GetCardID : NULL;
}

const BYTE * CCasCardReader::GetInitialCbc(void)
{
	// Descrambler CBC Initial Value Ԃ
	return (m_hCasCard)? m_CasCardInfo.InitialCbc : NULL;
}

const BYTE * CCasCardReader::GetSystemKey(void)
{
	// Descrambling System Key Ԃ
	return (m_hCasCard)? m_CasCardInfo.SystemKey : NULL;
}

const BYTE * CCasCardReader::GetKsFromEcm(const BYTE *pEcmData, const DWORD dwEcmSize)
{
	static const BYTE abyEcmReceiveCmd[] = {0x90U, 0x34U, 0x00U, 0x00U};

	// uECM Receive Commandv
	if(!m_hCasCard){
		m_dwEcmError = EC_NO_ERROR;
		return NULL;
		}

	// ECMTCY`FbN
	if(!pEcmData || (dwEcmSize < 30UL) || (dwEcmSize > 256UL)){
		m_dwEcmError = EC_BADARGUMENT;
		return NULL;
		}

	// obt@
	DWORD dwRecvSize = 0UL;
	BYTE SendData[1024];
	BYTE RecvData[1024];
	::ZeroMemory(RecvData, sizeof(RecvData));

	// R}h\z
	::CopyMemory(SendData, abyEcmReceiveCmd, sizeof(abyEcmReceiveCmd));			// CLA, INS, P1, P2
	SendData[sizeof(abyEcmReceiveCmd)] = (BYTE)dwEcmSize;						// COMMAND DATA LENGTH
	::CopyMemory(&SendData[sizeof(abyEcmReceiveCmd) + 1], pEcmData, dwEcmSize);	// ECM
	SendData[sizeof(abyEcmReceiveCmd) + dwEcmSize + 1] = 0x00U;					// RESPONSE DATA LENGTH

	// J[hgUNVJn
	SCardTransaction transaction(m_hCasCard);

	// R}hM
	if(!TransmitCommand(m_hCasCard, SendData, sizeof(abyEcmReceiveCmd) + dwEcmSize + 2UL, RecvData, sizeof(RecvData), &dwRecvSize)){
		::ZeroMemory(&m_EcmStatus, sizeof(m_EcmStatus));
		m_dwEcmError = EC_TRANSMIT_ERROR;
		return NULL;
		}

	// TCY`FbN
	if(dwRecvSize != 25UL){
		::ZeroMemory(&m_EcmStatus, sizeof(m_EcmStatus));
		m_dwEcmError = EC_TRANSMIT_ERROR;
		return NULL;
		}	
	
	// X|X
	::CopyMemory(m_EcmStatus.KsData, &RecvData[6], sizeof(m_EcmStatus.KsData));

	// ^[R[h
	switch(((WORD)RecvData[4] << 8) | (WORD)RecvData[5]){
		// wς: 
		case 0x0200U :	// 㕥PPV
		case 0x0400U :	// OPPV
		case 0x0800U :	// eBA
		
		// w: vr[
		case 0x4480U :	// 㕥PPV
		case 0x4280U :	// OPPV
			m_dwEcmError = EC_NO_ERROR;
			return m_EcmStatus.KsData;
		
		// LȊO(s)
		default :
			m_dwEcmError = EC_NOT_CONTRACTED;
			return NULL;
		}
}

const DWORD CCasCardReader::GetEcmError(void)
{
	// ECMG[R[hԂ
	return m_dwEcmError;
}

void CCasCardReader::ResetCard(void)
{
	if(!m_hCasCard)return;

	// J[hZbg
	::SCardBeginTransaction(m_hCasCard);
	::SCardEndTransaction(m_hCasCard, SCARD_RESET_CARD);
}

const bool CCasCardReader::SendEmmSection(const BYTE *pEmmData, const DWORD dwEmmSize)
{
	static const BYTE abyEmmReceiveCmd[] = {0x90U, 0x36U, 0x00U, 0x00U};

	// uEMM Receive Commandv
	if(!m_hCasCard)return false;

	// EMMTCY`FbN
	if(!pEmmData || (dwEmmSize < 17UL) || (dwEmmSize > 263UL))return false;

	// obt@
	DWORD dwRecvSize = 0UL;
	BYTE SendData[1024];
	BYTE RecvData[1024];
	::ZeroMemory(RecvData, sizeof(RecvData));

	// R}h\z
	::CopyMemory(SendData, abyEmmReceiveCmd, sizeof(abyEmmReceiveCmd));			// CLA, INS, P1, P2
	SendData[sizeof(abyEmmReceiveCmd)] = (BYTE)dwEmmSize;						// COMMAND DATA LENGTH
	::CopyMemory(&SendData[sizeof(abyEmmReceiveCmd) + 1], pEmmData, dwEmmSize);	// EMM
	SendData[sizeof(abyEmmReceiveCmd) + dwEmmSize + 1] = 0x00U;					// RESPONSE DATA LENGTH

	// J[hgUNVJn
	SCardTransaction transaction(m_hCasCard);

	// R}hM
	if(!TransmitCommand(m_hCasCard, SendData, sizeof(abyEmmReceiveCmd) + dwEmmSize + 2UL, RecvData, sizeof(RecvData), &dwRecvSize))return false;

	// TCY`FbN
	if(dwRecvSize != 8UL)return false;
	
	// ^[R[h
	switch(((WORD)RecvData[4] << 8) | (WORD)RecvData[5]){
		case 0x2100U :	// I
			return true;

		case 0xA102U :	// ^p(^pOvgRԍ)
		case 0xA107U :	// ZLeBG[(EMMG[)
			return false;
		
		// LȊO(O)
		default :
			return false;
		}
}

const WORD CCasCardReader::GetPowerCtrlInfo(POWERCTRLINFO *pPowerCtrlInfo)
{
	BYTE abyReqPowerCtrlCmd[] = {0x90U, 0x80U, 0x00U, 0x00U, 0x01U, 0x00U, 0x00U};

	// ʓd擾
	if(!m_hCasCard)return 0U;

	// obt@
	DWORD dwRecvSize = 0UL;
	BYTE RecvData[1024];
	::ZeroMemory(RecvData, sizeof(RecvData));

	// J[hgUNVJn
	SCardTransaction transaction(m_hCasCard);

	// R}hM
	if(!TransmitCommand(m_hCasCard, abyReqPowerCtrlCmd, sizeof(abyReqPowerCtrlCmd), RecvData, sizeof(RecvData), &dwRecvSize))return 0U;

	// MTCY`FbN
	if(dwRecvSize != 20UL)return 0U;		

	// ^[R[h
	if((((WORD)RecvData[4] << 8) | (WORD)RecvData[5]) != 0x2100U)return 0U;

	// X|X
	const WORD wInfoNum = (WORD)RecvData[7] + 1U;
	if(!pPowerCtrlInfo)return wInfoNum;		// NULL̂Ƃ͕Kvȃobt@Ԃ

	CTsTime TsTime;

	for(WORD wIndex = 0U ; wIndex < wInfoNum ; wIndex++){
		// R}hݒ
		abyReqPowerCtrlCmd[5] = (BYTE)wIndex;

		// R}hM
		if(!TransmitCommand(m_hCasCard, abyReqPowerCtrlCmd, sizeof(abyReqPowerCtrlCmd), RecvData, sizeof(RecvData), &dwRecvSize))return 0U;

		// MTCY`FbN
		if(dwRecvSize != 20UL)return 0U;		

		// ^[R[h
		if((((WORD)RecvData[4] << 8) | (WORD)RecvData[5]) != 0x2100U)return 0U;
		
		// X|X
		TsTime.ClearTime();
		CAribTime::SplitAribMjd(((WORD)RecvData[9] << 8) | (WORD)RecvData[10], &TsTime.wYear, &TsTime.wMonth, &TsTime.wDay, &TsTime.wDayOfWeek);
		TsTime -= (DWORD)RecvData[11] * 24UL * 60UL * 60UL;
		pPowerCtrlInfo[wIndex].stStartTime = TsTime;
		TsTime += (DWORD)RecvData[12] * 24UL * 60UL * 60UL;
		pPowerCtrlInfo[wIndex].stEndTime = TsTime;
		pPowerCtrlInfo[wIndex].byDurTime = RecvData[13];
		pPowerCtrlInfo[wIndex].wNtID = (((WORD)RecvData[14] << 8) | (WORD)RecvData[15]);
		pPowerCtrlInfo[wIndex].wTsID = (((WORD)RecvData[16] << 8) | (WORD)RecvData[17]);
		}

	return wInfoNum;
}

CCasCardReader::CCasCardReader(IBonObject *pOwner)
	: CBonObject(pOwner)
	, m_ScardContext(NULL)
	, m_hCasCard(NULL)
	, m_dwEcmError(EC_NOT_OPEN)
{
	// ԏ
	::ZeroMemory(&m_CasCardInfo, sizeof(m_CasCardInfo));
	::ZeroMemory(&m_EcmStatus, sizeof(m_EcmStatus));

	// \[X}l[WReLXgm
	::SCardEstablishContext(SCARD_SCOPE_USER, NULL, NULL, &m_ScardContext);
}

CCasCardReader::~CCasCardReader(void)
{
	CloseCard();

	// \[X}l[WReLXg̊J
	if(m_ScardContext)::SCardReleaseContext(m_ScardContext);
}

const bool CCasCardReader::EnumCasCardReader(void) const
{
	// J[h[_񋓂
	DWORD dwBuffSize = 0UL;
	
	// obt@TCY擾
	if(::SCardListReaders(m_ScardContext, NULL, NULL, &dwBuffSize) != SCARD_S_SUCCESS)return false;

	// obt@m
	TCHAR *pszReaders = new TCHAR[dwBuffSize];

	// J[h[_
	if(::SCardListReaders(m_ScardContext, NULL, pszReaders, &dwBuffSize) != SCARD_S_SUCCESS){
		delete [] pszReaders;
		return false;
		}

	// J[h[_ۑ
	TCHAR *lpszCurReader = pszReaders;
	m_CardReaderArray.clear();
			
	while(*lpszCurReader){
		m_CardReaderArray.push_back(lpszCurReader);
		lpszCurReader += m_CardReaderArray.back().GetLength() + 1UL;
		}

	delete[] pszReaders;

	return (m_CardReaderArray.size())? true : false;
}

const bool CCasCardReader::TransmitCommand(const SCARDHANDLE hCasCard, const BYTE *pSendData, const DWORD dwSendSize, BYTE *pRecvData, const DWORD dwMaxRecv, DWORD *pdwRecvSize) const
{
	DWORD dwRecvSize = dwMaxRecv;

	// f[^M
	DWORD dwReturn = ::SCardTransmit(hCasCard, SCARD_PCI_T1, pSendData, dwSendSize, NULL, pRecvData, &dwRecvSize);
	
	// MTCYi[
	if(pdwRecvSize)*pdwRecvSize = dwRecvSize;

	return (dwReturn == SCARD_S_SUCCESS)? true : false;
}

const bool CCasCardReader::InitialSetting(const SCARDHANDLE hCasCard) const
{
	static const BYTE abyInitSettingCmd[] = {0x90U, 0x30U, 0x00U, 0x00U, 0x00U};

	// uInitial Setting Conditions Commandv
	if(!hCasCard)return false;

	// obt@
	DWORD dwRecvSize = 0UL;
	BYTE RecvData[1024];
	::ZeroMemory(RecvData, sizeof(RecvData));

	// J[hgUNVJn
	SCardTransaction transaction(hCasCard);

	// R}hM
	if(!TransmitCommand(hCasCard, abyInitSettingCmd, sizeof(abyInitSettingCmd), RecvData, sizeof(RecvData), &dwRecvSize))return false;

	// MTCY`FbN
	if(dwRecvSize < 57UL)return false;		

	// X|X
	::CopyMemory(m_CasCardInfo.GetCardID, &RecvData[8],   6UL);	// +8	Card ID
	::CopyMemory(m_CasCardInfo.SystemKey,  &RecvData[16], 32UL);	// +16	Descrambling system key
	::CopyMemory(m_CasCardInfo.InitialCbc, &RecvData[48],  8UL);	// +48	Descrambler CBC initial value

	// ECMXe[^X
	::ZeroMemory(&m_EcmStatus, sizeof(m_EcmStatus));

	return true;
}
