//////////////////////////////////////////////////////////////////////////
// BonSDK CAS card reader library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// CasCardReader.h: CASJ[h[_NX
//
/////////////////////////////////////////////////////////////////////////////

#pragma once

#include <Vector>
#include <String>
#include <WinScard.h>


/////////////////////////////////////////////////////////////////////////////
// ėpX}[gJ[h[_p CASJ[hC^tF[XNX
/////////////////////////////////////////////////////////////////////////////

class CCasCardReader :	public CBonObject,
						public IHalDevice,
						public IHalCasCard
{
public:
// CBonObject
	DECLARE_IBONOBJECT(CCasCardReader)

// IHalDevice
	virtual const BONGUID GetDeviceType(void) const;
	virtual const CBonString GetDeviceName() const;
	virtual const DWORD GetTotalDeviceNum(void) const;
	virtual const DWORD GetActiveDeviceNum(void) const;

// IHalCasCard
	virtual const bool OpenCard(void);
	virtual void CloseCard(void);
	virtual const bool IsInteractive(void);
	virtual const BYTE * GetCardID(void);
	virtual const BYTE * GetInitialCbc(void);
	virtual const BYTE * GetSystemKey(void);
	virtual const BYTE * GetKsFromEcm(const BYTE *pEcmData, const DWORD dwEcmSize);
	virtual const DWORD GetEcmError(void);
	virtual void ResetCard(void);
	virtual const bool SendEmmSection(const BYTE *pEmmData, const DWORD dwEmmSize);
	virtual const WORD GetPowerCtrlInfo(POWERCTRLINFO *pPowerCtrlInfo);

// CCasCardReader
	CCasCardReader(IBonObject *pOwner);
	virtual ~CCasCardReader(void);
	
protected:
	const bool EnumCasCardReader(void) const;

private:
	const bool TransmitCommand(const SCARDHANDLE hCasCard, const BYTE *pSendData, const DWORD dwSendSize, BYTE *pRecvData, const DWORD dwMaxRecv, DWORD *pdwRecvSize = NULL) const;
	const bool InitialSetting(const SCARDHANDLE hCasCard) const;

	struct TAG_CASCARDINFO
	{
		BYTE GetCardID[6];		// Card ID
		BYTE SystemKey[32];		// Descrambling system key
		BYTE InitialCbc[8];		// Descrambler CBC initial value
	};

	struct TAG_ECMSTATUS
	{
		BYTE KsData[16];		// Ks Odd + Even	
	};

	SCARDCONTEXT m_ScardContext;
	SCARDHANDLE m_hCasCard;

	mutable std::vector<CBonString> m_CardReaderArray;
	
	mutable TAG_CASCARDINFO m_CasCardInfo;
	mutable TAG_ECMSTATUS m_EcmStatus;
	mutable DWORD m_dwEcmError;
};
