//////////////////////////////////////////////////////////////////////////
// BonSDK Media decoder library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// ProgAnalyzer.h: vOAiCUfR[_
//
/////////////////////////////////////////////////////////////////////////////

#pragma once


#include "TsTableIf.h"
#include "TsPidMapper.h"
#include <Vector>


/////////////////////////////////////////////////////////////////////////////
// vOAiCUfR[_
/////////////////////////////////////////////////////////////////////////////
// Input	#0	: ITsPacket		TSpPbg
// Output	#0	: ITsPacket		TSpPbg
/////////////////////////////////////////////////////////////////////////////

class CProgAnalyzer :	public CMediaDecoder,
						public IProgAnalyzer,
						protected IPatTable::IHandler,
						protected IPmtTable::IHandler,
						protected INitTable::IHandler,
						protected ISdtTable::IHandler,
						protected ITotTable::IHandler,
						protected IEitTable::IHandler
{
public:
// CBonObject
	DECLARE_IBONOBJECT(CProgAnalyzer)

// IProgAnalyzer
	virtual const WORD GetNetworkID(void) const;
	virtual const WORD GetTsID(void) const;
	virtual const WORD GetProgramIDFromPID(const WORD wPID) const;
	virtual const WORD GetPmtPIDFromProgramID(const WORD wProgramID) const;
	virtual const DWORD GetProgramNum(void) const;
	virtual const WORD GetProgramID(const DWORD dwIndex = 0UL) const;
	virtual const WORD GetProgramVideoPID(const WORD wProgramID = 0x0000U) const;
	virtual const WORD GetProgramAudioPID(const WORD wProgramID = 0x0000U) const;
	virtual const WORD GetProgramPcrPID(const WORD wProgramID = 0x0000U) const;
	virtual const CBonString GetNetworkName() const;
	virtual const CBonString GetServiceName(const WORD wProgramID = 0x0000U) const;
	virtual const bool GetStreamTime(SYSTEMTIME *pStreamTime) const;
	virtual const WORD GetCurrentProgramID(void) const;
	virtual const bool SetCurrentProgramID(const WORD wProgramID = 0x0000U);
	virtual const IEpgItem * GetCurrentEpgInfo(const WORD wProgramID = 0x0000U) const;
	virtual const IEpgItem * GetNextEpgInfo(const WORD wProgramID = 0x0000U) const;
	
// CProgAnalyzer
	CProgAnalyzer(IBonObject *pOwner, IMediaDataHeap *pHeap = new CMediaDataHeap());
	virtual ~CProgAnalyzer(void);
	
protected:
// CMediaDecoder
	virtual const bool OnPlay(void);
	virtual const bool OnReset(void);
	virtual const bool OnInputMedia(const IMediaData *pMediaData, const DWORD dwInputIndex = 0UL);

// IPatTable::IHandler
	virtual void OnPatTable(const IPatTable *pPatTable);

// IPmtTable::IHandler
	virtual void OnPmtTable(const IPmtTable *pPmtTable);

// NitTable::IHandler
	virtual void OnNitTable(const INitTable *pNitTable);

// ISdtTable::IHandler
	virtual void OnSdtTable(const ISdtTable *pSdtTable);

// ITotTable::IHandler
	virtual void OnTotTable(const ITotTable *pTotTable);

// EitTable::IHandler
	virtual void OnEitTable(const IEitTable *pEitTable, const IEitItem *pEitItem);

private:
// CProgAnalyzer
	struct PROGRAM_ITEM
	{
		bool bAvailable;			// LtO
		WORD wPmtPID;				// PmtPID
		WORD wProgramID;			// vOID
		WORD wVideoPID;				// fPID
		WORD wAudioPID;				// PID
		WORD wPcrPID;				// PCRPID
		CBonString ServiceName;		// T[rX
		IEpgItem *pCurEpgItem;		// EPG
		IEpgItem *pNextEpgItem;		// EPG
	};

	typedef std::vector<PROGRAM_ITEM> ProgramList;

	struct PROGRAM_CONTEXT
	{
		WORD wNetworkID;			// lbg[NID
		WORD wTsID;					// TS ID
		WORD wCurProgramID;			// ݃vOID
		CBonString NetworkName;		// lbg[N
		ProgramList ProgList;
	};

	void ResetPidMap(void);
	PROGRAM_ITEM * FindProgramItem(WORD wProgramID) const;

	CTsPidMapper m_TsPidMapper;
	PROGRAM_CONTEXT m_ProgContext;
};
