//////////////////////////////////////////////////////////////////////////
// BonSDK Media decoder library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// TsTable.cpp: TSe[uNX
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "BonSDK.h"
#include <TChar.h>
#include "TsTable.h"


/////////////////////////////////////////////////////////////////////////////
// [0x00] PATe[uNX
/////////////////////////////////////////////////////////////////////////////

void CPatTable::Reset(void)
{
	// ԂNA
	m_wTsID = 0x0000U;
	
	m_NitPIDArray.clear();
	m_PmtPIDArray.clear();

	CPsiTableBase::Reset();
}

const WORD CPatTable::GetTsID(void) const
{
	// TS IDԂ
	return m_wTsID;
}

const WORD CPatTable::GetNitPID(const DWORD dwIndex) const
{
	// NITPIDԂ
	return (dwIndex < m_NitPIDArray.size())? m_NitPIDArray[dwIndex].wPID : TS_INVALID_PID;
}

const DWORD CPatTable::GetNitNum(void) const
{
	// NIT̐Ԃ
	return m_NitPIDArray.size();
}

const WORD CPatTable::GetPmtPID(const DWORD dwIndex) const
{
	// PMTPIDԂ
	return (dwIndex < m_PmtPIDArray.size())? m_PmtPIDArray[dwIndex].wPID : TS_INVALID_PID;
}

const WORD CPatTable::GetProgramID(const DWORD dwIndex) const
{
	// Program Number IDԂ
	return (dwIndex < m_PmtPIDArray.size())? m_PmtPIDArray[dwIndex].wProgramID : 0x0000U;
}

const DWORD CPatTable::GetProgramNum(void) const
{
	// PMT̐Ԃ
	return m_PmtPIDArray.size();
}

const WORD CPatTable::GetPmtPIDFromPID(const WORD wPID) const
{
	for (DWORD dwIndex = 0; dwIndex < m_PmtPIDArray.size(); dwIndex++)
	{
		if (m_PmtPIDArray[dwIndex].wPID == wPID)
		{
			return m_PmtPIDArray[dwIndex].wProgramID;
		}
	}
	return 0;
}

CPatTable::CPatTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_pHandler(dynamic_cast<IPatTable::IHandler *>(pOwner))
	, m_wTsID(0UL)
{
	// Ȃ
}

CPatTable::~CPatTable(void)
{
	// Ȃ
}

const bool CPatTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	const WORD wDataSize = pNewSection->GetPayloadSize();
	const BYTE *pHexData = pNewSection->GetPayloadData();

	if(wDataSize % 4U)return false;										// e[ũTCYs
	if(pNewSection->GetTableID() != IPatTable::TABLE_ID)return false;	// e[uIDs

	// PIDNA
	m_NitPIDArray.clear();
	m_PmtPIDArray.clear();

	TAG_PATITEM PatItem;

	// e[u͂
	m_wTsID = pNewSection->GetTableIdExtension();

	CBonString trace = CBonString::Format(TEXT("------- PAT Table -------\r\nTS ID = %1!04X!\r\n"), m_wTsID);

	for(WORD wPos = 0U ; wPos < wDataSize ; wPos += 4U, pHexData += 4){
		PatItem.wProgramID	= ((WORD)pHexData[0] << 8) | (WORD)pHexData[1];				// +1,2
		PatItem.wPID		= ((WORD)(pHexData[2] & 0x1FU) << 8) | (WORD)pHexData[3];	// +3,4

		if(!PatItem.wProgramID){
			// NITPID
			trace.AppendFormat(TEXT("NIT #%1!u! [ID:%2!04X!][PID:%3!04X!]\r\n"), m_NitPIDArray.size(), PatItem.wProgramID, PatItem.wPID);
			m_NitPIDArray.push_back(PatItem);
			}
		else{
			// PMTPID
			trace.AppendFormat(TEXT("PMT #%1!u! [ID:%2!04X!][PID:%3!04X!]\r\n"), m_PmtPIDArray.size(), PatItem.wProgramID, PatItem.wPID);
			m_PmtPIDArray.push_back(PatItem);
			}
		}

	BON_TRACE(TraceType_Debug, trace);

	// nh֒ʒm
	if(m_pHandler)m_pHandler->OnPatTable(this);

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// [0x01] CATe[uNX
/////////////////////////////////////////////////////////////////////////////

void CCatTable::Reset(void)
{
	// ԂNA
	m_DescBlock.Reset();

	CPsiTableBase::Reset();
}

const IDescBlock * CCatTable::GetCatDesc(void) const
{
	// Lqq̈Ԃ
	return &m_DescBlock;
}

CCatTable::CCatTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_DescBlock(NULL)
	, m_pHandler(dynamic_cast<ICatTable::IHandler *>(pOwner))
{
	// Ȃ
}

CCatTable::~CCatTable(void)
{
	// Ȃ
}

const bool CCatTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	if(pNewSection->GetTableID() != ICatTable::TABLE_ID)return false;	// e[uIDs
	
	// e[u͂
	m_DescBlock.SetData(pNewSection->GetPayloadData(), pNewSection->GetPayloadSize());
	m_DescBlock.ParseDescs();

	CBonString trace = TEXT("------- CAT Table -------\r\n");

	for(DWORD dwIndex = 0UL ; dwIndex < m_DescBlock.GetDescNum() ; dwIndex++){
		const IDescBase *pDescBase = m_DescBlock.GetDescByIndex(dwIndex);
		trace.AppendFormat(TEXT("[%1!lu!] TAG = 0x%2!02X! LEN = %3!lu!\r\n"), dwIndex, pDescBase->GetTag(), pDescBase->GetLength());
		}

	BON_TRACE(TraceType_Debug, trace);

	// nh֒ʒm
	if(m_pHandler)m_pHandler->OnCatTable(this);

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// [0x02] PMTe[uNX
/////////////////////////////////////////////////////////////////////////////

void CPmtTable::Reset(void)
{
	// ԂNA
	m_wProgramID = 0x0000U;
	m_wPcrPID = TS_INVALID_PID;
	m_DescBlock.Reset();

	for(DWORD dwIndex = 0UL ; dwIndex < m_EsInfoArray.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_EsInfoArray[dwIndex].pDescBlock);
		}

	m_EsInfoArray.clear();
	
	CPsiTableBase::Reset();
}

const WORD CPmtTable::GetProgramID(void) const
{
	// ԑgԍIDԂ
	return m_wProgramID;
}

const WORD CPmtTable::GetPcrPID(void) const
{
	// PCR PIDԂ
	return m_wPcrPID;
}

const IDescBlock * CPmtTable::GetPmtDesc(void) const
{
	// Lqq̈1Ԃ
	return &m_DescBlock;
}

const DWORD CPmtTable::GetEsNum(void) const
{
	// G^Xg[Ԃ
	return m_EsInfoArray.size();
}

const BYTE CPmtTable::GetStreamTypeID(const DWORD dwIndex) const
{
	// Xg[`IDԂ
	return (dwIndex < m_EsInfoArray.size())? m_EsInfoArray[dwIndex].byStreamTypeID : 0x00U;
}

const WORD CPmtTable::GetEsPID(const DWORD dwIndex) const
{
	// G^PIDԂ
	return (dwIndex < m_EsInfoArray.size())? m_EsInfoArray[dwIndex].wEsPID : TS_INVALID_PID;
}

const IDescBlock * CPmtTable::GetEsDesc(const DWORD dwIndex) const
{
	// Lqq̈2Ԃ
	return (dwIndex < m_EsInfoArray.size())? m_EsInfoArray[dwIndex].pDescBlock : NULL;
}

CPmtTable::CPmtTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_wProgramID(0x0000U)
	, m_wPcrPID(TS_INVALID_PID)
	, m_DescBlock(NULL)
	, m_pHandler(dynamic_cast<IPmtTable::IHandler *>(pOwner))
{
	// Ȃ
}

CPmtTable::~CPmtTable(void)
{
	// CX^XJ
	Reset();
}

const bool CPmtTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	const WORD wDataSize = pNewSection->GetPayloadSize();
	const BYTE *pHexData = pNewSection->GetPayloadData();

	if(pNewSection->GetTableID() != IPmtTable::TABLE_ID)return false;	// e[uIDs
	
	// ESNA
	for(DWORD dwIndex = 0UL ; dwIndex < m_EsInfoArray.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_EsInfoArray[dwIndex].pDescBlock);
		}

	m_EsInfoArray.clear();
	
	// e[u͂
	m_wProgramID = pNewSection->GetTableIdExtension();
	m_wPcrPID = ((WORD)(pHexData[0] & 0x1FU) << 8) | (WORD)pHexData[1];

	WORD wDescLen = ((WORD)(pHexData[2] & 0x0FU) << 8) | (WORD)pHexData[3];
	m_DescBlock.SetData(&pHexData[4], wDescLen);
	m_DescBlock.ParseDescs();

	CBonString trace = CBonString::Format(
		TEXT("------- PMT Table -------\r\nProgram Number ID = %1!04X!\r\nPCR PID = %2!04X!\r\nECM PID = %3!04X!\r\nDescriptor Field:\r\n"),
		m_wProgramID, m_wPcrPID,
		(m_DescBlock.GetDescByTag(CCaMethodDesc::DESC_TAG))?
		dynamic_cast<const CCaMethodDesc *>(m_DescBlock.GetDescByTag(CCaMethodDesc::DESC_TAG))->GetCaPID() : TS_INVALID_PID);

	for(DWORD dwIndex = 0UL ; dwIndex < m_DescBlock.GetDescNum() ; dwIndex++){
		const IDescBase *pDescBase = m_DescBlock.GetDescByIndex(dwIndex);
		trace.AppendFormat(TEXT("    [%1!lu!] TAG = 0x%2!02X! LEN = %3!lu!\r\n"),
			dwIndex, pDescBase->GetTag(), pDescBase->GetLength());
		}

	// Xg[̊JnʒuvZ
	WORD wPos = wDescLen + 4U;
	TAG_PMTITEM PmtItem;

	// ES
	while(wPos < wDataSize){
		PmtItem.byStreamTypeID = pHexData[wPos + 0U];													// +0
		PmtItem.wEsPID = ((WORD)(pHexData[wPos + 1U] & 0x1FU) << 8) | (WORD)pHexData[wPos + 2U];		// +1,2	
		wDescLen = ((WORD)(pHexData[wPos + 3U] & 0x0FU) << 8) | (WORD)pHexData[wPos + 4U];				// +3,4

		// LqqubN
		PmtItem.pDescBlock = dynamic_cast<CDescBlock *>(CDescBlock::CreateInstance(NULL));
		PmtItem.pDescBlock->SetData(&pHexData[wPos + 5U], wDescLen);
		PmtItem.pDescBlock->ParseDescs();

		CBonString trace = CBonString::Format(TEXT("[%1!u!] Stream Type ID = %2!02X!  PID = %3!04X!\r\n"),
			m_EsInfoArray.size(), PmtItem.byStreamTypeID, PmtItem.wEsPID);		

		for(DWORD dwIndex = 0UL ; dwIndex < PmtItem.pDescBlock->GetDescNum() ; dwIndex++){
			const IDescBase *pDescBase = PmtItem.pDescBlock->GetDescByIndex(dwIndex);
			trace.AppendFormat(TEXT("    [%1!lu!] TAG = 0x%2!02X! LEN = %3!lu!\r\n"),
				dwIndex, pDescBase->GetTag(), pDescBase->GetLength());
			}

		// e[uɒǉ
		m_EsInfoArray.push_back(PmtItem);
		wPos += (wDescLen + 5U);
		}

	BON_TRACE(TraceType_Debug, trace);

	// nh֒ʒm
	if(m_pHandler)m_pHandler->OnPmtTable(this);

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// [0x40] NITe[uNX
/////////////////////////////////////////////////////////////////////////////

void CNitTable::Reset(void)
{
	// ԂNA
	m_wNetworkID = 0x0000U;
	m_NwDescBlock.Reset();

	for(DWORD dwIndex = 0UL ; dwIndex < m_TsInfoArray.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_TsInfoArray[dwIndex].pDescBlock);
		}

	m_TsInfoArray.clear();
	
	CPsiTableBase::Reset();
}

const WORD CNitTable::GetNetworkID(void) const
{
	// lbg[NIDԂ
	return m_wNetworkID;
}

const IDescBlock * CNitTable::GetNetworkDesc(void) const
{
	// lbg[NLqq̈Ԃ
	return (m_NwDescBlock.GetSize())? &m_NwDescBlock : NULL;
}

const DWORD CNitTable::GetTsNum(void) const
{
	// TS񐔂Ԃ
	return m_TsInfoArray.size();
}

const BYTE CNitTable::GetTsID(const DWORD dwIndex) const
{
	// TS IDԂ
	return (dwIndex < m_TsInfoArray.size())? m_TsInfoArray[dwIndex].wTsID : 0x0000U;
}

const BYTE CNitTable::GetOrgNetworkID(const DWORD dwIndex) const
{
	// IWilbg[NIDԂ
	return (dwIndex < m_TsInfoArray.size())? m_TsInfoArray[dwIndex].wOrgNetworkID : 0x0000U;
}

const IDescBlock * CNitTable::GetTsDesc(const DWORD dwIndex) const
{
	// lbg[NLqq̈Ԃ
	return (dwIndex < m_TsInfoArray.size())? m_TsInfoArray[dwIndex].pDescBlock : NULL;
}

CNitTable::CNitTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_wNetworkID(0x0000U)
	, m_NwDescBlock(NULL)
	, m_pHandler(dynamic_cast<INitTable::IHandler *>(pOwner))
{
	// Ȃ
}

CNitTable::~CNitTable(void)
{
	// CX^XJ
	Reset();
}

const bool CNitTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	const WORD wDataSize = pNewSection->GetPayloadSize();
	const BYTE *pHexData = pNewSection->GetPayloadData();

	if(pNewSection->GetTableID() != INitTable::TABLE_ID)return false;	// e[uIDs
	
	// TSNA
	for(DWORD dwIndex = 0UL ; dwIndex < m_TsInfoArray.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_TsInfoArray[dwIndex].pDescBlock);
		}

	m_TsInfoArray.clear();
	
	// e[u͂
	m_wNetworkID = pNewSection->GetTableIdExtension();

	WORD wDescLen = ((WORD)(pHexData[0] & 0x0FU) << 8) | (WORD)pHexData[1];
	m_NwDescBlock.SetData(&pHexData[2], wDescLen);
	m_NwDescBlock.ParseDescs();

	CBonString trace = CBonString::Format(TEXT("------- NIT Table -------\r\nNetwork ID = %1!04X!\r\n"), m_wNetworkID);
	
	CBonString temp(TEXT("Unknown"));
	if(m_NwDescBlock.GetDescByTag(INetworkNameDesc::DESC_TAG)){
		temp = dynamic_cast<const INetworkNameDesc *>(m_NwDescBlock.GetDescByTag(INetworkNameDesc::DESC_TAG))->GetNetworkName();
		}
	trace.AppendFormat(TEXT("Network Name = %s\r\n"), temp.GetString());

	trace += TEXT("Descriptor Field:\r\n");
	for(DWORD dwIndex = 0UL ; dwIndex < m_NwDescBlock.GetDescNum() ; dwIndex++){
		const IDescBase *pDescBase = m_NwDescBlock.GetDescByIndex(dwIndex);
		trace.AppendFormat(TEXT("    [%1!lu!] TAG = 0x%2!02X! LEN = %3!lu!\r\n"), dwIndex, pDescBase->GetTag(), pDescBase->GetLength());
		}

	// TS̊JnʒuvZ
	WORD wPos = wDescLen + 4U;
	TAG_NITITEM NitItem;

	// TS
	while(wPos < wDataSize){
		NitItem.wTsID = ((WORD)pHexData[wPos + 0U] << 8) | (WORD)pHexData[wPos + 1U];
		NitItem.wOrgNetworkID = ((WORD)pHexData[wPos + 2U] << 8) | (WORD)pHexData[wPos + 3U];

		wDescLen = ((WORD)(pHexData[wPos + 4U] & 0x0FU) << 8) | (WORD)pHexData[wPos + 5U];

		// LqqubN
		NitItem.pDescBlock = dynamic_cast<CDescBlock *>(CDescBlock::CreateInstance(NULL));
		NitItem.pDescBlock->SetData(&pHexData[wPos + 6U], wDescLen);
		NitItem.pDescBlock->ParseDescs();

		trace.AppendFormat(TEXT("[%1!u!] TS ID = %2!04X!  Original Network ID = %3!04X!\r\n    Descriptor Field:\r\n"),
			m_TsInfoArray.size(), NitItem.wTsID, NitItem.wOrgNetworkID);		

		for(DWORD dwIndex = 0UL ; dwIndex < NitItem.pDescBlock->GetDescNum() ; dwIndex++){
			const IDescBase *pDescBase = NitItem.pDescBlock->GetDescByIndex(dwIndex);
			trace.AppendFormat(TEXT("        [%1!lu!] TAG = 0x%2!02X! LEN = %3!lu!\r\n"), dwIndex, pDescBase->GetTag(), pDescBase->GetLength());
			}

		// e[uɒǉ
		m_TsInfoArray.push_back(NitItem);
		wPos += (wDescLen + 6U);
		}

	BON_TRACE(TraceType_Debug, trace);

	// nh֒ʒm
	if(m_pHandler)m_pHandler->OnNitTable(this);

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// [0x42] SDTe[uNX
/////////////////////////////////////////////////////////////////////////////

void CSdtTable::Reset(void)
{
	// ԂNA
	m_wTsID = 0x0000U;
	m_wOrgNetworkID = 0x0000U;

	for(DWORD dwIndex = 0UL ; dwIndex != m_SdtItemArray.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_SdtItemArray[dwIndex].pDescBlock);
		}

	m_SdtItemArray.clear();
	
	CPsiTableBase::Reset();
}

const WORD CSdtTable::GetTsID(void) const
{
	// TS IDԂ
	return m_wTsID;
}

const WORD CSdtTable::GetOrgNetworkID(void) const
{
	// IWilbg[NIDԂ
	return m_wOrgNetworkID;
}

const DWORD CSdtTable::GetServiceNum(void) const
{
	// T[rXԂ
	return m_SdtItemArray.size();
}

const WORD CSdtTable::GetServiceID(const DWORD dwIndex) const
{
	// T[rXIDԂ
	return (dwIndex < m_SdtItemArray.size())? m_SdtItemArray[dwIndex].wServiceID : 0x0000U;
}

const bool CSdtTable::IsEitSchedule(const WORD wServiceID) const
{
	// EITXPW[tOԂ
	const DWORD dwIndex = GetIndexByServiceID(wServiceID);
	return (dwIndex < m_SdtItemArray.size())? m_SdtItemArray[wServiceID].bIsEitSchedule : false;
}

const bool CSdtTable::IsEitPresentFollowing(const WORD wServiceID) const
{
	// EIT/tOԂ
	const DWORD dwIndex = GetIndexByServiceID(wServiceID);
	return (dwIndex < m_SdtItemArray.size())? m_SdtItemArray[dwIndex].IsEitPresentFollowing : false;
}

const BYTE CSdtTable::GetRunningStatus(const WORD wServiceID) const
{
	// isԂԂ
	const DWORD dwIndex = GetIndexByServiceID(wServiceID);
	return (dwIndex < m_SdtItemArray.size())? m_SdtItemArray[dwIndex].byRunningStatus : ISdtTable::RST_UNDEFINED;
}

const bool CSdtTable::IsScrambling(const WORD wServiceID) const
{
	// XNutOԂ
	const DWORD dwIndex = GetIndexByServiceID(wServiceID);
	return (dwIndex < m_SdtItemArray.size())? m_SdtItemArray[dwIndex].bIsScrambling : false;
}

const IDescBlock * CSdtTable::GetServiceDesc(const WORD wServiceID) const
{
	// Lqq̈Ԃ
	const DWORD dwIndex = GetIndexByServiceID(wServiceID);
	return (dwIndex < m_SdtItemArray.size())? m_SdtItemArray[dwIndex].pDescBlock : NULL;
}

CSdtTable::CSdtTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_wTsID(0x0000U)
	, m_wOrgNetworkID(0x0000U)
	, m_pHandler(dynamic_cast<ISdtTable::IHandler *>(pOwner))
{
	// Ȃ
}

CSdtTable::~CSdtTable(void)
{
	// CX^XJ
	Reset();
}

const bool CSdtTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	const WORD wDataSize = pNewSection->GetPayloadSize();
	const BYTE *pHexData = pNewSection->GetPayloadData();

	if(pNewSection->GetTableID() != ISdtTable::TABLE_ID)return false;	// e[uIDs
	
	// T[rXNA
	for(DWORD dwIndex = 0UL ; dwIndex != m_SdtItemArray.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_SdtItemArray[dwIndex].pDescBlock);
		}

	m_SdtItemArray.clear();
	
	// e[u͂
	m_wTsID = pNewSection->GetTableIdExtension();
	m_wOrgNetworkID = ((WORD)pHexData[0] << 8) | (WORD)pHexData[1];

	CBonString trace = CBonString::Format(
		TEXT("------- SDT Table -------\r\nTS ID = %1!04X!\r\nOriginal Network ID = %2!04X!\r\n"), m_wTsID, m_wOrgNetworkID);

	WORD wPos = 3U;
	TAG_SDTITEM SdtItem;
	
	// T[rX
	while(wPos < wDataSize){
		SdtItem.wServiceID = ((WORD)pHexData[wPos + 0U] << 8) | (WORD)pHexData[wPos + 1U];
		SdtItem.bIsEitSchedule = (pHexData[wPos + 2U] & 0x02U)? true : false;
		SdtItem.IsEitPresentFollowing = (pHexData[wPos + 2U] & 0x01U)? true : false;
		SdtItem.byRunningStatus = (pHexData[wPos + 3U]) >> 5;
		SdtItem.bIsScrambling = (pHexData[wPos + 3U] & 0x10U)? true : false;

		// LqqubN
		const WORD wDescLen = ((WORD)(pHexData[wPos + 3U] & 0x0FU) << 8) | (WORD)pHexData[wPos + 4U];
		SdtItem.pDescBlock = dynamic_cast<CDescBlock *>(CDescBlock::CreateInstance(NULL));
		SdtItem.pDescBlock->SetData(&pHexData[wPos + 5U], wDescLen);
		SdtItem.pDescBlock->ParseDescs();

		// e[uɒǉ
		m_SdtItemArray.push_back(SdtItem);
		wPos += (wDescLen + 5U);

		// ȉfobOp
		trace.AppendFormat(TEXT("\r\n[%1!u!] Service ID = %2!04X!\r\n    Eit Schedule = %3!u!\r\n    Eit Present Following = %4!u!\r\n    Running Status = %5!u!\r\n    Scrambling = %6!u!\r\n"),
			m_SdtItemArray.size() - 1U, SdtItem.wServiceID, SdtItem.bIsEitSchedule, SdtItem.IsEitPresentFollowing, SdtItem.byRunningStatus, SdtItem.bIsScrambling);		

		CBonString temp(TEXT("Unknown"));
		if(SdtItem.pDescBlock->GetDescByTag(IServiceDesc::DESC_TAG)){
			temp = dynamic_cast<const IServiceDesc *>(SdtItem.pDescBlock->GetDescByTag(IServiceDesc::DESC_TAG))->GetProviderName();
			}
		trace.AppendFormat(TEXT("    Provider Name = %1!s!\r\n"), temp.GetString());

		temp = TEXT("Unknown");
		if(SdtItem.pDescBlock->GetDescByTag(IServiceDesc::DESC_TAG)){
			temp = dynamic_cast<const IServiceDesc *>(SdtItem.pDescBlock->GetDescByTag(IServiceDesc::DESC_TAG))->GetServiceName();
			}
		trace.AppendFormat(TEXT("    Service Name = %1!s!\r\n"), temp.GetString());

		for(DWORD dwIndex = 0UL ; dwIndex < SdtItem.pDescBlock->GetDescNum() ; dwIndex++){
			const IDescBase *pDescBase = SdtItem.pDescBlock->GetDescByIndex(dwIndex);
			trace.AppendFormat(TEXT("        [%1!lu!] TAG = 0x%2!02X! LEN = %3!lu!\r\n"),
				dwIndex, pDescBase->GetTag(), pDescBase->GetLength());
			}
		}

	BON_TRACE(TraceType_Debug, trace);

	// nh֒ʒm
	if(m_pHandler)m_pHandler->OnSdtTable(this);

	return true;
}

const DWORD CSdtTable::GetIndexByServiceID(const WORD wServiceID) const
{
	// T[rXIDCfbNX擾
	DWORD dwIndex;
	
	for(dwIndex = 0UL ; dwIndex < m_SdtItemArray.size() ; dwIndex++){
		if(m_SdtItemArray[dwIndex].wServiceID == wServiceID)break;
		}

	return dwIndex;
}


/////////////////////////////////////////////////////////////////////////////
// [0x4E-0x6F] EITe[uNX
/////////////////////////////////////////////////////////////////////////////

void CEitTable::Reset(void)
{
	// ԂNA
	m_EitItem.ClearSize();
	
	CPsiTableBase::Reset();
}

const WORD CEitTable::GetServiceID(void) const
{
	// T[rXIDԂ
	return m_LastSection.GetTableIdExtension();
}

const WORD CEitTable::GetTsID(void) const
{
	// TS IDԂ
	if(m_LastSection.GetPayloadSize() < 6U)return 0x0000U;

	const BYTE *pHexData = m_LastSection.GetPayloadData();
	return ((WORD)pHexData[0] << 8) | (WORD)pHexData[1];
}

const WORD CEitTable::GetOrgNetworkID(void) const
{
	// Original Network IDԂ
	if(m_LastSection.GetPayloadSize() < 6U)return 0x0000U;

	const BYTE *pHexData = m_LastSection.GetPayloadData();
	return ((WORD)pHexData[2] << 8) | (WORD)pHexData[3];
}

const BYTE CEitTable::GetTableID(void) const
{
	// e[uIDԂ
	return m_LastSection.GetTableID();
}

const BYTE CEitTable::GetLastTableID(void) const
{
	// ŏIe[uIDԂ
	if(m_LastSection.GetPayloadSize() < 6U)return 0x00U;

	const BYTE *pHexData = m_LastSection.GetPayloadData();
	return pHexData[5];
}

const BYTE CEitTable::GetSectionNumber(void) const
{
	// ZNVԍԂ
	return m_LastSection.GetSectionNumber();
}

const BYTE CEitTable::GetLastSectionNumber(void) const
{
	// ŏIZNVԍԂ
	return m_LastSection.GetLastSectionNumber();
}

const BYTE CEitTable::GetSegmentLastSectionNumber(void) const
{
	// ZOgŏIZNVԍԂ
	if(m_LastSection.GetPayloadSize() < 6U)return 0x00U;

	const BYTE *pHexData = m_LastSection.GetPayloadData();
	return pHexData[4];
}

CEitTable::CEitTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_pHandler(dynamic_cast<IEitTable::IHandler *>(pOwner))
	, m_EitItem(NULL)
{
	// Ȃ
}

CEitTable::~CEitTable(void)
{
	// Ȃ
}

void CEitTable::OnPsiSection(const IPsiSectionParser *pPsiSectionParser, const IPsiSection *pPsiSection)
{
	// CPsiTableBase̎I[o[Ch
	if(m_LastSection.CompareSection(pPsiSection)){
		// ZNVXVꂽ
		OnTableUpdate(pPsiSection, &m_LastSection);
		}
}

const bool CEitTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	const WORD wDataSize = pNewSection->GetPayloadSize();
	const BYTE *pHexData = pNewSection->GetPayloadData();

	if(pNewSection->GetTableID() < IEitTable::TABLE_ID_START)return false;	// e[uIDs
	if(pNewSection->GetTableID() > IEitTable::TABLE_ID_END)return false;	// e[uIDs
	if(wDataSize < 18U)return false;										// yC[hs

	// ZNVXgA
	m_LastSection.CopySection(pNewSection);

	// ACeɃnhɒʒm
	if(!m_pHandler)return true;
	
	for(WORD wPos = 6U ; (wPos < wDataSize) && ((wDataSize - wPos) >= 12U) ; ){
		// ACeTCYvZ
		const WORD wItemLen = (((WORD)(pHexData[wPos + 10] & 0x0FU) << 8) | (WORD)pHexData[wPos + 11]) + 12U;
		m_EitItem.SetData(&pHexData[wPos], wItemLen);
		wPos += wItemLen;
		
		// nh֒ʒm
		if(m_EitItem.ParseItem()){
			m_pHandler->OnEitTable(this, &m_EitItem);
			}
		}	

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// [0x73] TOTe[uNX
/////////////////////////////////////////////////////////////////////////////

void CTotTable::Reset(void)
{
	// ԂNA
	m_DateTime.ClearTime();
	m_DescBlock.Reset();
	
	CPsiTableBase::Reset();
}

const SYSTEMTIME & CTotTable::GetDateTime(void) const
{
	// ݓt/Ԃ
	return m_DateTime;
}

const IDescBlock * CTotTable::GetTotDesc(void) const
{
	// Lqq̈Ԃ
	return &m_DescBlock;
}

CTotTable::CTotTable(IBonObject *pOwner)
	: CPsiTableBase(pOwner)
	, m_DescBlock(NULL)
	, m_pHandler(dynamic_cast<ITotTable::IHandler *>(pOwner))
{
	// W`ZNVI
	m_SectionParser.SetExtSection(false);
}

CTotTable::~CTotTable(void)
{
	// Ȃ
}

const bool CTotTable::OnTableUpdate(const IPsiSection *pNewSection, const IPsiSection *pLastSection)
{
	const WORD wDataSize = pNewSection->GetPayloadSize();
	const BYTE *pHexData = pNewSection->GetPayloadData();

	if(pNewSection->GetTableID() != ITotTable::TABLE_ID)return false;	// e[uIDs

	// e[u͂
	m_DateTime.SetAribTime(pHexData);
	
	const WORD wDescLen = ((WORD)(pHexData[5] & 0x0FU) << 8) | (WORD)pHexData[6];
	m_DescBlock.SetData(&pHexData[7], wDescLen);
	m_DescBlock.ParseDescs();

	// nh֒ʒm
	if(m_pHandler)m_pHandler->OnTotTable(this);

	return true;
}
